import argparse
import builtins
import json
import os
import tempfile
from enum import Enum, auto
from itertools import product, zip_longest
from pathlib import Path

import pytest
import yaml

from vembrane import __version__, errors
from vembrane.backend.base import Backend
from vembrane.common import create_reader
from vembrane.modules import fhir, filter, sort, structured, table, tag

FILTER_CASES = Path(__file__).parent.joinpath("testcases/filter")
TABLE_CASES = Path(__file__).parent.joinpath("testcases/table")
TAG_CASES = Path(__file__).parent.joinpath("testcases/tag")
ANNOTATE_CASES = Path(__file__).parent.joinpath("testcases/annotate")
STRUCTURED_CASES = Path(__file__).parent.joinpath("testcases/structured")
FHIR_CASES = Path(__file__).parent.joinpath("testcases/fhir")
SORT_CASES = Path(__file__).parent.joinpath("testcases/sort")


def test_version():
    assert __version__ != "unknown"


def idfn(val):
    if isinstance(val, os.PathLike):
        return "-".join(os.path.normpath(val).split(os.sep)[-2:])


class Context(Enum):
    FILE = auto()
    STATEMENT = auto()
    NONE = auto()
    ERROR = auto()


@pytest.mark.parametrize(
    "testcase,backend,context",
    product(
        (
            case_path.joinpath(d)
            for case_path in [
                FILTER_CASES,
                TABLE_CASES,
                TAG_CASES,
                ANNOTATE_CASES,
                STRUCTURED_CASES,
                FHIR_CASES,
                SORT_CASES,
            ]
            for d in os.listdir(case_path)
            if not d.startswith(".")
        ),
        (Backend.pysam, Backend.cyvcf2),
        (Context.NONE, Context.FILE, Context.STATEMENT, Context.ERROR),
    ),
    ids=idfn,
)
def test_command(testcase: os.PathLike, backend: Backend, context: Context | None):
    path = testcase
    vcf_path = path.joinpath("test.vcf")

    with open(path.joinpath("config.yaml")) as config_fp:
        config = yaml.load(config_fp, Loader=yaml.FullLoader)

    # emulate command-line command setup to use argparse afterwards
    cmd = config["function"]
    command = parse_command_config(cmd, config, vcf_path)

    match context:
        case Context.STATEMENT:
            command.extend(["--context", "import random; dummy_func = lambda x: x"])
        case Context.FILE:
            command.extend(["--context-file", "tests/resources/context.py"])
        case Context.ERROR:
            if "raises" not in config:
                command.extend(["--context-file", "tests/resources/context_error.py"])
                config["raises"] = "VembraneError"
            else:
                pytest.skip(
                    "Test expects other error, skipping context error injection."
                )
        case Context.NONE:
            if config.get("needs_context"):
                pytest.skip("Test needs context, skipping.")

    parser = construct_parser()

    try:
        args, unknown = parser.parse_known_args(command)
    # we have to catch and compare such exceptions here, because they
    # do not cause a SystemExit like the .execute() instances below
    except Exception as e:
        assert type(e).__name__ == config["raises"]
        return

    if "raises" in config:
        exception = config["raises"]
        try:
            exception = getattr(errors, exception)
        except AttributeError:
            exception = getattr(builtins, exception)

        with pytest.raises(SystemExit), pytest.raises(exception):
            if args.command == "filter":
                filter.execute(args)
            elif args.command == "table":
                table.execute(args)
            elif args.command == "tag":
                tag.execute(args)
            elif args.command == "structured":
                structured.execute(args)
            elif args.command == "fhir":
                fhir.execute(args)
            elif args.command == "sort":
                sort.execute(args)
            else:
                raise AssertionError from None
    else:
        with tempfile.NamedTemporaryFile(mode="w+t") as tmp_out:
            args.output = tmp_out.name
            args.backend = backend
            if (
                args.command == "filter"
                or args.command == "tag"
                or args.command == "sort"
            ):
                match args.command:
                    case "filter":
                        module = filter
                    case "tag":
                        module = tag
                    case "sort":
                        module = sort
                expected = str(path.joinpath("expected.vcf"))
                module.execute(args)
                with create_reader(tmp_out.name, backend=backend) as vcf_actual:
                    with create_reader(expected, backend=backend) as vcf_expected:
                        for r1, r2 in zip_longest(vcf_actual, vcf_expected):
                            assert r1 == r2

                        assert (
                            vcf_actual.header.get_generic("vembraneVersion")
                            == __version__
                        )
            elif args.command == "table":
                expected = str(path.joinpath("expected.tsv"))
                table.execute(args)
                t_out = "".join(
                    line for line in tmp_out if not line.startswith("##vembrane")
                )
                with open(expected) as e:
                    e_out = e.read()
                assert t_out == e_out
            elif args.command == "structured" or args.command == "fhir":
                expected = (path / "expected").with_suffix(f".{config['output_fmt']}")

                if args.command == "structured":
                    structured.execute(args)
                else:
                    fhir.execute(args)

                with open(expected) as e:
                    if config["output_fmt"] == "jsonl":
                        t_out = [json.loads(line) for line in tmp_out]
                        e_out = [json.loads(line) for line in e]
                    elif config["output_fmt"] == "json":
                        t_out = json.load(tmp_out)
                        e_out = json.load(e)
                    elif config["output_fmt"] == "yaml":
                        t_out = yaml.safe_load(tmp_out)
                        e_out = yaml.safe_load(e)
                assert t_out == e_out
            else:
                assert args.command in {
                    "filter",
                    "table",
                    "tag",
                    "structured",
                    "fhir",
                    "sort",
                }, "Unknown subcommand"


def construct_parser():
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers(
        dest="command",
        description="valid subcommands",
        required=True,
    )
    filter.add_subcommmand(subparsers)
    table.add_subcommmand(subparsers)
    tag.add_subcommand(subparsers)
    structured.add_subcommand(subparsers)
    fhir.add_subcommand(subparsers)
    sort.add_subcommand(subparsers)
    return parser


def parse_command_config(cmd, config, vcf_path):
    if cmd in ("filter", "table"):
        command = [cmd, config["expression"], str(vcf_path)]
    elif cmd == "tag":
        command = [cmd]
        tags = config["tags"]
        for name, expr in tags.items():
            command += ["--tag", f"{name}={expr}"]
        command.append(str(vcf_path))
    elif cmd == "structured":
        template_path = vcf_path.parent / "template.yte.yaml"
        command = [cmd, str(template_path), str(vcf_path)]
    elif cmd == "fhir":
        command = [cmd, str(vcf_path), config["sample"], config["assembly"]]
        del config["sample"]
        del config["assembly"]
    elif cmd == "sort":
        command = [cmd, config["expression"], str(vcf_path)]
        del config["expression"]
    else:
        raise ValueError(f"Unknown subcommand {config['function']}")
    for key in config:
        if key == "function":
            continue
        if isinstance(config[key], str):
            command.append(config_key_to_arg(key))
            command.append(config[key])
        elif isinstance(config[key], bool):
            if config[key]:
                command.append(config_key_to_arg(key))
        elif isinstance(config[key], (int, float)):
            command.append(config_key_to_arg(key))
            command.append(str(config[key]))
        else:
            if isinstance(config[key], dict):
                for k, v in config[key].items():
                    command.append(config_key_to_arg(key))
                    command.append(f"{k}={v}")
            else:
                command.append(config_key_to_arg(key))
                for argument in config[key]:
                    command.append(argument)
    return command


def config_key_to_arg(key: str) -> str:
    """Convert a config key to an argument name."""
    return f"--{key.replace('_', '-')}"



import ast
import sys
from types import CodeType
from typing import Any

from .ann_types import ANN_TYPER, NA, MoreThanOneAltAlleleError, NvFloat, NvInt
from .backend.base import VCFHeader, VCFRecord, VCFRecordFormats, VCFRecordInfo
from .common import get_annotation_keys, split_annotation_entry
from .errors import (
    MalformedAnnotationError,
    NonBoolTypeError,
    UnknownAnnotationError,
    VembraneError,
)
from .globals import _explicit_clear, custom_functions, default_allowed_globals
from .sequence_ontology import _C, SequenceOntology


class NoValueDict:
    def __contains__(self, item) -> bool:
        try:
            value = self[item]  # type: ignore
        except KeyError:
            return False
        return value is not NA


class DefaultGet:
    def get(self, item, default=NA):
        v = self[item]
        if v is not NA:
            return v
        else:
            return default


class Annotation(NoValueDict, DefaultGet):
    def __init__(self, ann_key: str, header: VCFHeader) -> None:
        self._record_idx = -1
        self._record: VCFRecord | None = None
        self._annotation_data: list[str] = []
        self._data: dict[str, Any] = {}
        self._ann_key = ann_key
        annotation_keys = get_annotation_keys(header, ann_key)
        self._ann_conv = {
            entry.name: (ann_idx, entry.convert)
            for ann_idx, entry in enumerate(map(ANN_TYPER.get_entry, annotation_keys))
        }

    def update(self, record_idx: int, record: VCFRecord, annotation: str):
        self._record_idx = record_idx
        self._record = record
        self._data.clear()
        self._annotation_data = split_annotation_entry(annotation)

    def get_record_annotations(self, idx: int, record: VCFRecord) -> list[str]:
        # if the expression contains a reference to the ANN field
        # get all annotations from the record.info field
        # (or supply an empty ANN value if the record has no ANN field)
        annotations = record.info[self._ann_key]
        if annotations is NA:
            num_ann_entries = len(self._ann_conv.keys())
            empty = "|" * num_ann_entries
            print(
                f"No ANN field found in record {idx}, "
                f"replacing with NAs (i.e. 'ANN={empty}')",
                file=sys.stderr,
            )
            annotations = [empty]
        return annotations

    def set_record_annotations(self, record: VCFRecord, annotations: list[str]) -> None:
        # If the given new annotations are all empty and ANN was NA before,
        # we keep the NA.
        # The check 'ann == len(ann) * "|"' is the fastest way to check if a string
        # is just "||||||", according to https://stackoverflow.com/a/14321721.
        # Even faster than regex.
        if (
            all(ann == len(ann) * "|" for ann in annotations)
            and record.info[self._ann_key] is NA
        ):
            return
        record.info[self._ann_key] = annotations

    def __getitem__(self, item):
        try:
            return self._data[item]
        except KeyError:
            try:
                ann_idx, convert = self._ann_conv[item]
            except KeyError as ke2:
                raise UnknownAnnotationError(
                    self._record,
                    item,
                ) from ke2
            if ann_idx >= len(self._annotation_data):
                raise MalformedAnnotationError(
                    self._record,
                    item,
                    ann_idx,
                ) from None
            raw_value = self._annotation_data[ann_idx].strip()
            value = self._data[item] = convert(raw_value)
            return value

    def __bool__(self):
        """Evaluates to False if no annotation is present in the header."""
        return bool(self._ann_conv)


UNSET = object()


class WrapFloat32Visitor(ast.NodeTransformer):
    def visit_Constant(self, node):
        from ctypes import c_float

        if not isinstance(node.value, float):
            return node

        return ast.Constant(c_float(node.value).value)


class Environment(dict):
    def __init__(
        self,
        ann_key: str,
        header: VCFHeader,
        auxiliary: dict[str, set[str]] | None = None,
        ontology: SequenceOntology | None = None,
        allowed_globals: dict[str, Any] | None = None,
        auxiliary_globals: dict[str, Any] | None = None,
    ) -> None:
        if allowed_globals is None:
            allowed_globals = default_allowed_globals
        if auxiliary is None:
            auxiliary = {}
        if ann_key == "ANN":
            self.item_is_ann = lambda item: item == "ANN"
        else:
            self.item_is_ann = lambda item: item == "ANN" or item == ann_key
        self._annotation: Annotation = Annotation(ann_key, header)
        self._globals: dict[str, Any] = dict(allowed_globals)
        # We use self + self.func as a closure.
        self._globals.update(custom_functions(self))
        self._globals["SAMPLES"] = list(header.samples)

        # REF/ALT alleles are cached separately to raise "MoreThanOneAltAllele"
        # only if ALT (but not REF) is accessed (and ALT has multiple entries).
        self._alleles = None

        self._getters = {
            "AUX": self._get_aux,
            "SO": self._get_ontology,
            "CHROM": self._get_chrom,
            "POS": self._get_pos,
            "END": self._get_end,
            "ID": self._get_id,
            "ALT": self._get_alt,
            "REF": self._get_ref,
            "QUAL": self._get_qual,
            "FILTER": self._get_filter,
            "INFO": self._get_info,
            "FORMAT": self._get_format,
            "INDEX": self._get_index,
        }

        self.update(**_explicit_clear)

        if auxiliary_globals:
            common_names = [
                name
                for name in auxiliary_globals
                if name in self._getters or name in self._globals
            ]
            if common_names:
                raise VembraneError(
                    "Context definition may not overwrite names defined by vembrane. "
                    f"Conflicting: {','.join(common_names)}."
                )
            self._globals.update(auxiliary_globals)

        # vembrane only supports bi-allelic records (i.e. one REF, one ALT allele).
        # Hence, for all fields in the header whith `number in {"A", "R"}`
        # we check if there is indeed only 1 value ("A") or 2 values ("R")
        # and abort otherwise.
        # Then, in the case of `number == "A"`, the value tuples only have one entry,
        # so that the value can be accessed directly and need not be accessed via
        # an index operation.

        # always explicitly set "Number" for certain fields
        # which get special pysam treatment:
        # - `FORMAT["GT"]` is always parsed as a list of integer values
        # self._numbers.get("FORMAT", {})["GT"] = "."

        # At the moment, only INFO and FORMAT records are checked
        self._empty_globals = {name: UNSET for name in self._getters}
        self.record: VCFRecord = None  # type: ignore
        self.idx: int = -1
        self.aux = auxiliary
        self.so = ontology
        if ontology:
            _C.__dict__["ontology"] = ontology

    def update_from_record(self, idx: int, record: VCFRecord):
        self.idx = idx
        self.record = record
        self._globals.update(self._empty_globals)
        self._alleles = None

    def update_data(self, data):
        self._globals["DATA"] = data

    def update_annotation(self, annotation):
        self._annotation.update(self.idx, self.record, annotation)

    def get_record_annotations(self, idx: int, record: VCFRecord) -> list[str]:
        return self._annotation.get_record_annotations(idx, record)

    def _get_chrom(self) -> str:
        value = self.record.contig
        self._globals["CHROM"] = value
        return value

    def _get_pos(self) -> int:
        value = self.record.position
        self._globals["POS"] = value
        return value

    def _get_end(self) -> NvInt:
        value = self.record.end
        self._globals["END"] = value
        return value

    def _get_id(self) -> str | None:
        value = self.record.id
        self._globals["ID"] = value
        return value

    def _get_alleles(self) -> tuple[str, ...]:
        alleles = self._alleles
        if not alleles:
            alleles = self._alleles = self.record.alleles  # type: ignore
        return alleles

    def _get_ref(self) -> str:
        alleles = self._get_alleles()
        value = alleles[0]
        self._globals["REF"] = value
        return value

    def _get_alt(self) -> str:
        alleles = self._get_alleles()
        if len(alleles) > 2:
            raise MoreThanOneAltAlleleError
        value = alleles[1] if len(alleles) == 2 else NA
        self._globals["ALT"] = value
        return value

    def _get_qual(self) -> NvFloat:
        value: NvFloat = NA if self.record.quality is None else self.record.quality
        self._globals["QUAL"] = value
        return value

    def _get_filter(self) -> list[str]:
        value = list(self.record.filter)
        self._globals["FILTER"] = value
        return value

    def _get_info(self) -> VCFRecordInfo:
        value = self.record.info
        self._globals["INFO"] = value
        return value

    def _get_format(self) -> VCFRecordFormats:
        value = self.record.formats
        self._globals["FORMAT"] = value
        return value

    def __getitem__(self, item):
        if self.item_is_ann(item):
            # we always allow to access the annotation key with ANN
            return self._annotation
        value = self._globals[item]
        if value is UNSET:
            value = self._getters[item]()
        return value

    def _get_index(self) -> int:
        self._globals["INDEX"] = self.idx
        return self.idx

    def _get_aux(self) -> dict[str, set[str]]:
        self._globals["AUX"] = self.aux
        return self.aux

    def _get_ontology(self) -> SequenceOntology:
        if not self.so:
            self.so = SequenceOntology.default()
        self._globals["SO"] = self.so
        return self.so  # type: ignore


class SourceEnvironment(Environment):
    def __init__(
        self,
        source: str,
        ann_key: str,
        header: VCFHeader,
        auxiliary: dict[str, set[str]] | None = None,
        ontology: SequenceOntology | None = None,
        allowed_globals: dict[str, Any] | None = None,
        auxiliary_globals: dict[str, Any] | None = None,
    ):
        super().__init__(
            ann_key, header, auxiliary, ontology, allowed_globals, auxiliary_globals
        )

        self._has_ann: bool = any(
            hasattr(node, "id")
            and isinstance(node, ast.Name)
            and self.item_is_ann(node.id)
            for node in ast.walk(ast.parse(source))
        )

        # The VCF specification only allows 32bit floats.
        # Comparisons such as `INFO["some_float"] > CONST` may yield unexpected results,
        # because `some_float` is a 32bit float while `CONST` is a python 64bit float.
        # Example: `c_float(0.6) = 0.6000000238418579 > 0.6`.
        # We can work around this by wrapping user provided floats in `ctypes.c_float`,
        # which should follow the same IEEE 754 specification as the VCF spec, as most C
        # compilers should follow this standard (https://stackoverflow.com/a/17904539):

        # parse the expression, obtaining an AST
        source_ast = ast.parse(source, mode="eval")

        # wrap each float constant in numpy.float32
        source_ast = WrapFloat32Visitor().visit(source_ast)

        # housekeeping
        source_ast = ast.fix_missing_locations(source_ast)

        # compile the now-fixed code-tree
        self.compiled: CodeType = compile(source_ast, filename="<string>", mode="eval")

    def expression_annotations(self) -> bool:
        return self._has_ann


class FuncWrappedExpressionEnvironment(SourceEnvironment):
    def __init__(
        self,
        expression: str,
        ann_key: str,
        header: VCFHeader,
        auxiliary: dict[str, set[str]] | None = None,
        ontology: SequenceOntology | None = None,
        allowed_globals: dict[str, Any] | None = None,
        auxiliary_globals: dict[str, Any] | None = None,
        evaluation_function_template: str = "lambda: {expression}",
    ) -> None:
        func_str: str = evaluation_function_template.format(expression=expression)
        super().__init__(
            func_str,
            ann_key,
            header,
            auxiliary,
            ontology,
            allowed_globals,
            auxiliary_globals,
        )
        self._func = eval(self.compiled, self, {})

    def is_true(self, annotation: str = "") -> bool:
        if self._has_ann:
            self.update_annotation(annotation)
        keep = self._func()
        if not isinstance(keep, bool):
            raise NonBoolTypeError(keep)
        return keep

    def table_row(self, annotation: str = "") -> tuple:
        if self._has_ann:
            self.update_annotation(annotation)
        return self._func()



import argparse

from vembrane.common import HumanReadableDefaultsFormatter

from . import __version__
from .modules import annotate, fhir, filter, sort, structured, table, tag


def main():
    parser = argparse.ArgumentParser(
        description="VCF/BCF filter and manipulation tool "
        "leveraging Python expressions.",
        formatter_class=HumanReadableDefaultsFormatter,
    )

    parser.add_argument(
        "--version",
        action="version",
        version=f"%(prog)s {__version__}",
    )
    subparsers = parser.add_subparsers(
        dest="command",
        description="valid subcommands",
        required=True,
    )
    filter.add_subcommmand(subparsers)
    table.add_subcommmand(subparsers)
    annotate.add_subcommmand(subparsers)
    tag.add_subcommand(subparsers)
    structured.add_subcommand(subparsers)
    fhir.add_subcommand(subparsers)
    sort.add_subcommand(subparsers)

    args = parser.parse_args()
    if args.command == "filter":
        filter.execute(args)
    elif args.command == "table":
        table.execute(args)
    elif args.command == "annotate":
        annotate.execute(args)
    elif args.command == "tag":
        tag.execute(args)
    elif args.command == "structured":
        structured.execute(args)
    elif args.command == "fhir":
        fhir.execute(args)
    elif args.command == "sort":
        sort.execute(args)
    else:
        raise ValueError(f"Unknown subcommand {args.command}")



from abc import abstractmethod, abstractproperty
from enum import Enum
from typing import List, Optional, Tuple

from ..ann_types import NA
from ..errors import UnknownSampleError


class Backend(Enum):
    pysam = 0
    cyvcf2 = 1

    def __str__(self):
        return self.name.lower()

    def __repr__(self):
        return str(self)

    @staticmethod
    def from_string(s):
        try:
            return Backend[s]
        except KeyError:
            return s


class VCFRecordInfo:
    __slots__ = ()

    @abstractmethod
    def __contains__(self, item):
        raise NotImplementedError

    @abstractmethod
    def __getitem__(self, item):
        raise NotImplementedError

    @abstractmethod
    def __setitem__(self, key, value):
        raise NotImplementedError

    @abstractmethod
    def keys(self):
        raise NotImplementedError

    def get(self, key, default=None):
        if key in self:
            return self[key]
        return default


class VCFRecordSamples:
    pass


class NoValueDict:
    def __contains__(self, item):
        try:
            value = self[item]
        except KeyError:
            return False
        return value is not NA


class DefaultGet:
    def get(self, item, default=NA):
        v = self[item]
        if v is not NA:
            return v
        else:
            return default


class VCFRecordFormat(NoValueDict):
    @abstractmethod
    def __setitem__(self, key, value):
        raise NotImplementedError

    def get(self, sample, default=None):
        try:
            return self[sample]
        except UnknownSampleError:
            return default

    @abstractmethod
    def keys(self):
        raise NotImplementedError

    def __repr__(self):
        return str(
            {
                sample: self.__getitem__(sample)
                for i, sample in enumerate(self._header.samples)
            },
        )


class VCFRecordFilter:
    __slots__ = ()

    @abstractmethod
    def add(self, tag: str):
        raise NotImplementedError

    @abstractmethod
    def __iter__(self):
        raise NotImplementedError


class VCFRecord:
    __slots__ = ("_raw_record", "record_idx", "_header")

    @abstractmethod
    def __init__(self, record, record_idx: int, header):
        self._raw_record = record
        self.record_idx = record_idx
        self._header = header

    @abstractproperty
    def contig(self) -> str:
        raise NotImplementedError

    def chrom(self) -> str:
        return self.contig

    @abstractproperty
    def position(self) -> int:
        raise NotImplementedError

    def start(self) -> int:
        return self.position

    @abstractproperty
    def stop(self) -> int:
        raise NotImplementedError

    @abstractproperty
    def id(self) -> str:
        raise NotImplementedError

    @property
    def alleles(self) -> Tuple[str, ...]:
        return self.reference_allele, *self.alt_alleles

    @abstractproperty
    def reference_allele(self) -> str:
        raise NotImplementedError

    @abstractproperty
    def alt_alleles(self) -> Tuple[str]:
        raise NotImplementedError

    @abstractproperty
    def quality(self) -> float:
        raise NotImplementedError

    @abstractproperty
    def filter(self) -> VCFRecordFilter:
        raise NotImplementedError

    @abstractproperty
    def info(self) -> VCFRecordInfo:
        raise NotImplementedError

    @abstractproperty
    def formats(self) -> "VCFRecordFormats":
        raise NotImplementedError

    @abstractproperty
    def header(self) -> "VCFHeader":
        raise NotImplementedError

    @property
    def is_bnd_record(self) -> bool:
        return "SVTYPE" in self.info and self.info.get("SVTYPE", None) == "BND"

    @property
    def is_sv_record(self) -> bool:
        is_sv = self.info.get("SVTYPE", "") in [
            "DEL",
            "INS",
            "DUP",
            "CNV",
            "INV",
            "INDEL",
            "BND",
            "TRA",
        ] or any(
            alt
            in [
                "<DEL>",
                "<DEL:ME>",
                "<INS>",
                "<INS:ME>",
                "<DUP>",
                "<DUP:TANDEM>",
                "<CNV>",
                "<CNV:TR>",
                "<INV>",
            ]
            for alt in self.alt_alleles
        )
        return is_sv

    @property
    def end(self):
        if self.is_bnd_record:
            return NA
        else:
            # record.stop is pysams unified approach to get the end position of
            # a variant. It either considers the alt allele or the END field,
            # depending on the record. Stop is 0-based, but for consistency
            # with POS we convert into 1-based. Since for 1-based coordinates,
            # the expectation is that END is inclusive instead of exclusive
            # (as it is with 0-based), we do not need to add 1.
            return self.stop

    @abstractmethod
    def __repr__(self):
        return self._raw_record.__str__()

    @abstractmethod
    def __str__(self):
        return self._raw_record.__str__()

    def __eq__(self, other: object):
        if not isinstance(other, VCFRecord):
            return NotImplemented
        return all(
            (
                self.contig == other.contig,
                self.id == other.id,
                self.alleles == other.alleles,
                self.position == other.position,
                self.quality == other.quality,
                set(self.filter) == set(other.filter),
                # because we might end up comparing NA == NA which is False by design,
                # we replace those with None here for the equality check,
                # i.e. None == None is True
                all(
                    (self.info.get(key) or None) == (other.info.get(key) or None)
                    for key in self.header.infos  # type: ignore
                ),
                all(
                    (self.formats.get(key) or None) == (other.formats.get(key) or None)
                    for key in self.header.formats  # type: ignore
                ),
            ),
        )


class VCFRecordFormats(NoValueDict):
    @abstractmethod
    def __init__(
        self,
        record: VCFRecord,
    ):
        raise NotImplementedError

    @abstractmethod
    def keys(self):
        raise NotImplementedError

    def get(self, key: str, default=None):
        if key not in self:
            return default


class VCFReader:
    __slots__ = ("_file",)

    @abstractmethod
    def __init__(self, filename: str):
        raise NotImplementedError

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def close(self):
        self._file.close()

    def __iter__(self):
        return self

    @property
    def file(self):
        return self._file

    @abstractmethod
    def __next__(self) -> VCFRecord:
        raise NotImplementedError

    @property
    def header(self):
        return self._header

    @abstractmethod
    def reset(self):
        raise NotImplementedError


class VCFHeader:
    @abstractmethod
    def __init__(self, reader: VCFReader):
        raise NotImplementedError

    @abstractmethod
    def __iter__(self):
        raise NotImplementedError

    @abstractmethod
    def __next__(self):
        raise NotImplementedError

    @abstractproperty
    def samples(self) -> List[str]:
        raise NotImplementedError

    @abstractproperty
    def filters(self) -> List[str]:
        raise NotImplementedError

    @abstractproperty
    def records(self):
        raise NotImplementedError

    @abstractproperty
    def infos(self) -> VCFRecordInfo:
        raise NotImplementedError

    @abstractproperty
    def formats(self) -> VCFRecordFormats:
        raise NotImplementedError

    @abstractmethod
    def add_meta(
        self,
        key: str,
        value: Optional[str] = None,
        items: Optional[List[Tuple[str, str]]] = None,
    ):
        raise NotImplementedError

    @abstractmethod
    def add_generic(self, key: str, value: str):
        raise NotImplementedError

    @abstractmethod
    def add_filter(self, id: str, description: str):
        raise NotImplementedError

    @abstractmethod
    def contains_generic(self, key: str):
        raise NotImplementedError


class VCFWriter:
    __slots__ = ("_file",)

    @abstractmethod
    def __init__(self, filename: str, fmt: str, template: VCFReader):
        raise NotImplementedError

    @abstractmethod
    def write(self, record: VCFRecord):
        raise NotImplementedError

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def close(self):
        self._file.close()



import argparse
import ast
import contextlib
import shlex
import sys
from collections import defaultdict
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Iterable, Iterator, TextIO, Type

from .backend.backend_cyvcf2 import Cyvcf2Reader, Cyvcf2Writer
from .backend.backend_pysam import PysamReader, PysamWriter
from .backend.base import Backend, VCFHeader, VCFReader, VCFRecord
from .errors import InvalidExpressionError, VembraneError
from .globals import _explicit_clear
from .sequence_ontology import SequenceOntology


def add_common_arguments(parser: argparse.ArgumentParser):
    parser.add_argument(
        "--annotation-key",
        "-k",
        metavar="FIELDNAME",
        default="ANN",
        help="The INFO key for the annotation field. "
        "This defaults to 'ANN', but tools might use other field names. "
        "For example, default VEP annotations can be parsed by setting 'CSQ' here.",
    )
    parser.add_argument(
        "--aux",
        "-a",
        nargs=1,
        action=AppendKeyValuePair,
        metavar="NAME=PATH",
        default={},
        help="Path to an auxiliary file containing a set of symbols",
    )
    parser.add_argument(
        "--context",
        help="Python statement defining a context for given Python expressions. "
        "Extends eventual definitions given via --context-file. "
        "Any global variables (or functions) become available in the Python "
        "expressions. Note that the code you pass here is not sandboxed and should "
        "be trusted. Carefully review any code you get from the internet or AI.",
    )
    parser.add_argument(
        "--context-file",
        help="Path to Python script defining a context for given Python expressions. "
        "Any global variables (or functions) become available in the Python "
        "expressions. Note that the code you pass here is not sandboxed and should "
        "be trusted. Carefully review any code you get from the internet or AI.",
    )
    parser.add_argument(
        "--ontology",
        nargs=1,
        default=None,
        metavar="PATH",
        help="Path to an ontology in OBO format. "
        "May be compressed with gzip, bzip2 and xz. "
        "Defaults to built-in ontology (from sequenceontology.org).",
    )
    parser.add_argument(
        "--overwrite-number-info",
        nargs=1,
        action=AppendKeyValuePair,
        metavar="FIELD=NUMBER",
        default={},
        help="Overwrite the number specification for INFO fields "
        "given in the VCF header. "
        "Example: `--overwrite-number cosmic_CNT=.`",
    )
    parser.add_argument(
        "--overwrite-number-format",
        nargs=1,
        action=AppendKeyValuePair,
        metavar="FIELD=NUMBER",
        default={},
        help="Overwrite the number specification for FORMAT fields "
        "given in the VCF header. "
        "Example: `--overwrite-number-format DP=2`",
    )
    parser.add_argument(
        "--backend",
        "-b",
        default="cyvcf2",
        type=Backend.from_string,
        choices=[Backend.cyvcf2, Backend.pysam],
        help="Set the backend library.",
    )


def check_expression(expression: str) -> str:
    if ".__" in expression:
        raise InvalidExpressionError(
            expression,
            "The expression must not contain '.__'",
        )
    try:
        tree = ast.parse(expression, mode="eval")
        if isinstance(tree.body, ast.BoolOp | ast.Compare):
            return expression
        else:
            # TODO possibly check for ast.Call, func return type
            return expression
    except SyntaxError as se:
        raise InvalidExpressionError(
            expression,
            "The expression has to be syntactically correct.",
        ) from se


def swap_quotes(s: str) -> str:
    return s.replace('"', '\\"').replace("'", '"').replace('\\"', "'")


def single_outer(s: str) -> bool:
    if '"' in s and "'" in s:
        return s.index('"') > s.index("'")
    elif '"' in s:
        return True
    elif "'" in s:
        return False
    return True


def normalize(s: str) -> str:
    return shlex.quote(swap_quotes(s) if not single_outer(s) else s)


def get_annotation_keys(header: VCFHeader, ann_key: str) -> list[str]:
    separator = "'"
    if header.contains_generic("VEP"):
        separator = ":"
    if h := header.infos.get(ann_key):
        try:
            return list(
                map(
                    str.strip,
                    h.get("Description").strip('"').split(separator)[1].split("|"),
                ),
            )
        except Exception as e:
            raise VembraneError(
                "Could not parse annotation keys from header. Has your VCF been "
                "properly annotated (e.g. with VEP or SnpEff)?"
            ) from e
    return []


def split_annotation_entry(entry: str) -> list[str]:
    return entry.split("|")


class BreakendEvent:
    __slots__ = ["name", "keep", "records", "keep_records", "mate_pair"]

    def __init__(self, name: str, mate_pair: bool = False) -> None:
        self.name = name
        self.records: list[VCFRecord] = []
        self.keep_records: list[bool] = []
        self.keep = False
        self.mate_pair = mate_pair

    def add(self, record: VCFRecord, keep_record: bool):
        self.records.append(record)
        self.keep_records.append(keep_record)
        self.keep |= keep_record

    def emit(self) -> Iterator[VCFRecord]:
        assert self.keep
        yield from self.records
        self.records = []
        self.keep_records = []
        # do not reset self.keep!

    def is_mate_pair(self) -> bool:
        return self.mate_pair

    def __str__(self) -> str:
        return self.name

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, other):
        return self.name == other.name


def mate_key(mates: Iterable[str | None]) -> str:
    return "__MATES: " + ",".join(sorted(m for m in mates if m is not None))


class AppendTagExpression(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        assert len(values) == 1
        if not hasattr(namespace, self.dest) or getattr(namespace, self.dest) is None:
            setattr(namespace, self.dest, {})
        value = values[0].strip()
        key, value = value.strip().split("=", 1)
        expr = check_expression(value)
        getattr(namespace, self.dest)[key.strip()] = expr


class AppendKeyValuePair(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        assert len(values) == 1
        if not hasattr(namespace, self.dest) or getattr(namespace, self.dest) is None:
            setattr(namespace, self.dest, {})
        value = values[0].strip()
        key, value = value.strip().split("=", 1)
        getattr(namespace, self.dest)[key.strip()] = value


def read_auxiliary(aux: dict[str, str]) -> dict[str, set[str]]:
    # read auxiliary files, split at any whitespace and store contents in a set
    def read_set(path: str) -> set[str]:
        with open(path) as f:
            return {line.rstrip() for line in f}

    return {name: read_set(contents) for name, contents in aux.items()}


def read_ontology(path: str | None) -> SequenceOntology | None:
    return SequenceOntology.from_obo(path) if path else None


@dataclass
class Context:
    context_file: str | None = None
    context_file_path: Path | None = None
    context_statement: str | None = None
    internal_context: str | None = None

    @classmethod
    def from_args(cls, args: argparse.Namespace) -> "Context":
        context_file = None
        context_statement = None
        internal_context = None
        context_file_path = None
        if path := args.context_file:
            context_file_path = Path(path).absolute()
            parent = context_file_path.parent
            internal_context = f"import sys\nsys.path.insert(0, '{parent}')\n"
            with open(path) as f:
                context_file = f.read()
        if statement := args.context:
            context_statement = statement
        return cls(
            context_file=context_file,
            context_file_path=context_file_path,
            context_statement=context_statement,
            internal_context=internal_context,
        )

    def get_globals(self) -> dict[str, Any]:
        globals_dict: dict[str, Any] = {}
        try:
            if self.internal_context:
                exec(self.internal_context, globals_dict)
            if self.context_file:
                exec(self.context_file, globals_dict)
            if self.context_statement:
                exec(self.context_statement, globals_dict)
        except Exception as e:
            raise VembraneError("Error while executing context: {e}") from e
        for name in _explicit_clear:
            if name in globals_dict:
                del globals_dict[name]
        return globals_dict


def create_reader(
    filename: str | Path,
    backend: Backend = Backend.pysam,
    overwrite_number=None,
) -> VCFReader:
    # GT should always be "."
    if overwrite_number is None:
        overwrite_number = defaultdict(dict)
    if backend == Backend.pysam:
        return PysamReader(filename, overwrite_number)
    elif backend == Backend.cyvcf2:
        return Cyvcf2Reader(filename, overwrite_number)
    else:
        raise ValueError(f"{backend} is not a known backend.")


def create_writer(
    filename: str | Path,
    fmt: str,
    template: VCFReader,
    backend: Backend = Backend.pysam,
):
    if backend == Backend.pysam:
        return PysamWriter(filename, fmt, template)
    elif backend == Backend.cyvcf2:
        return Cyvcf2Writer(filename, fmt, template)
    else:
        raise ValueError(f"{backend} is not a known backend.")


@contextlib.contextmanager
def smart_open(filename=None, *args, **kwargs) -> Iterator[TextIO]:
    fh = open(filename, *args, **kwargs) if filename and filename != "-" else sys.stdout

    try:
        yield fh
    finally:
        if fh is not sys.stdout:
            fh.close()


type Primitive = str | int | float | bool | None


# Inspired by https://stackoverflow.com/a/13793033
class Singleton(type):
    """A singleton metaclass: ensures that only one instance of a class exists."""

    _instances: dict[Type, Any] = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super().__call__(*args, **kwargs)
        return cls._instances[cls]


class HumanReadableDefaultsFormatter(argparse.HelpFormatter):
    """A custom argparse formatter that displays default values in a more
    human-readable way than the argparse's own defaults formatter which
    exposes Python types like dicts and None."""

    def _get_help_string(self, action):
        help_text = action.help
        if help_text is None:
            help_text = ""

        if action.default and action.default is not argparse.SUPPRESS:
            if isinstance(action.default, list):
                default_value = " ".join(map(str, action.default))
            elif isinstance(action.default, dict):
                default_value = " ".join(f"{k}={v}" for k, v in action.default.items())
            else:
                default_value = action.default
            help_text += f" (default: {default_value})"
        return help_text



import re
import sys
from itertools import chain, islice
from sys import stderr
from typing import Iterator

from .. import __version__
from ..ann_types import NA
from ..backend.base import VCFHeader, VCFReader, VCFRecord
from ..common import (
    AppendTagExpression,
    Context,
    HumanReadableDefaultsFormatter,
    add_common_arguments,
    check_expression,
    create_reader,
    create_writer,
    normalize,
    read_auxiliary,
    read_ontology,
    single_outer,
    swap_quotes,
)
from ..errors import (
    FilterAlreadyDefinedError,
    FilterTagNameInvalidError,
    VembraneError,
    handle_vembrane_error,
)
from ..representations import FuncWrappedExpressionEnvironment
from ..sequence_ontology import SequenceOntology
from .filter import DeprecatedAction


def add_subcommand(subparsers):
    parser = subparsers.add_parser(
        "tag",
        help="Add a flag to the FILTER field of VCF/BCF records without removing them "
        "(a non-destructive `filter`).",
        description="Flag records by adding a tag to their FILTER field "
        "based on one or more expressions. "
        "This is a non-destructive alternative to `filter`, "
        "as it keeps all records.",
        formatter_class=HumanReadableDefaultsFormatter,
    )
    parser.register("action", "deprecated", DeprecatedAction)
    parser.add_argument(
        "--tag",
        "-t",
        help="Tag records using the FILTER field.\n"
        "Note: tag names cannot contain `;` or whitespace and must not be '0'.\n"
        'Example: `--tag q_above_30="not (QUAL<=30)"`',
        nargs=1,
        metavar="TAG=EXPRESSION",
        action=AppendTagExpression,
        required=True,
    )
    parser.add_argument(
        "vcf",
        help="Path to the VCF/BCF file to be filtered. Defaults to '-' for stdin.",
        nargs="?",
        default="-",
    )
    parser.add_argument(
        "--output",
        "-o",
        default="-",
        help="Output file, if not specified, output is written to STDOUT.",
    )
    parser.add_argument(
        "--output-fmt",
        "-O",
        default="vcf",
        choices=["vcf", "bcf", "uncompressed-bcf"],
        help="Output format.",
    )
    parser.add_argument(
        "--tag-mode",
        "-m",
        default="pass",
        choices=["pass", "fail"],
        help="Set, whether to tag records that pass the tag expression(s), "
        "or records that fail them."
        "By default, vembrane tags records for which the tag expression(s) pass. "
        "This allows for descriptive tag names such as `q_at_least_30`, "
        "which would correspond to an expression `QUAL >= 30`. "
        "However, the VCF specification (`v4.4`) defines tags to be set when a "
        "filter expression is failed, so vembrane also offers the `fail` mode.",
    )
    add_common_arguments(parser)


def test_record(
    env: FuncWrappedExpressionEnvironment,
    idx: int,
    record: VCFRecord,
    ann_key: str,
) -> tuple[VCFRecord, bool]:
    env.update_from_record(idx, record)
    if env.expression_annotations():
        # if the expression contains a reference to the ANN field
        # get all annotations from the record.info field
        # (or supply an empty ANN value if the record has no ANN field)
        annotations = record.info[ann_key]
        if annotations is NA:
            num_ann_entries = len(env._annotation._ann_conv.keys())
            empty = "|" * num_ann_entries
            print(
                f"No ANN field found in record {idx}, "
                f"replacing with NAs (i.e. 'ANN={empty}')",
                file=sys.stderr,
            )
            annotations = [empty]

        #  … and check if the expression evaluates to true for any  of the annotations
        filtered = any(map(env.is_true, annotations))
        return record, filtered
    else:
        # otherwise, the annotations are irrelevant w.r.t. the expression,
        # so we can omit them
        return record, env.is_true()


def tag_vcf(
    vcf: VCFReader,
    expressions: dict[str, str],
    ann_key: str,
    auxiliary: dict[str, set[str]] | None = None,
    ontology: SequenceOntology | None = None,
    invert: bool = False,
    auxiliary_globals: dict[str, object] | None = None,
) -> Iterator[VCFRecord]:
    if auxiliary is None:
        auxiliary = {}

    # For each tag-expression pair, a different Environment must be used.
    envs = {
        tag: FuncWrappedExpressionEnvironment(
            expression,
            ann_key,
            vcf.header,
            auxiliary,
            ontology,
            auxiliary_globals=auxiliary_globals,
        )
        for tag, expression in expressions.items()
    }

    record: VCFRecord
    for idx, record in enumerate(vcf):
        for tag, env in envs.items():
            record, keep = test_record(env, idx, record, ann_key)
            if invert:
                keep = not keep
            if keep:
                record.filter.add(tag)
        yield record


def check_tag(tag: str):
    if re.search(r"^0$|[\s;]", tag):
        raise FilterTagNameInvalidError(tag)


@handle_vembrane_error
def execute(args) -> None:
    aux = read_auxiliary(args.aux)
    ontology = read_ontology(args.ontology)
    overwrite_number = {
        "INFO": dict(args.overwrite_number_info),
        "FORMAT": dict(args.overwrite_number_format),
    }

    with create_reader(
        args.vcf,
        backend=args.backend,
        overwrite_number=overwrite_number,
    ) as reader:
        header: VCFHeader = reader.header
        expressions = dict(args.tag)
        for tag, expr in expressions.items():
            for t, _rec in reader.header.filters.items():
                if t == tag:
                    e = FilterAlreadyDefinedError(tag)
                    print(e, file=stderr)
                    sys.exit(1)
            try:
                check_tag(tag)
            except VembraneError as ve:
                print(ve, file=stderr)
                sys.exit(1)
            expression = swap_quotes(expr) if single_outer(expr) else expr
            check_expression(expression)
            reader.header.add_filter(tag, expression)

        header.add_generic("vembraneVersion", __version__)
        header.add_generic(
            "vembraneCmd",
            "vembrane "
            + " ".join(normalize(arg) if " " in arg else arg for arg in sys.argv[1:]),
        )

        records = tag_vcf(
            reader,
            expressions,
            args.annotation_key,
            auxiliary=aux,
            ontology=ontology,
            invert=(args.tag_mode == "fail"),
            auxiliary_globals=Context.from_args(args).get_globals(),
        )

        first_record = list(islice(records, 1))

        records = chain(first_record, records)
        fmt = {"vcf": "", "bcf": "b", "uncompressed-bcf": "u"}[args.output_fmt]

        with create_writer(args.output, fmt, reader, backend=args.backend) as writer:
            for record in records:
                writer.write(record)



import sys
from typing import TYPE_CHECKING, Any, List, Self

if TYPE_CHECKING:
    from vembrane.backend.base import VCFRecord


class VembraneError(Exception):
    """Basic exception for errors raised by vembrane"""

    @classmethod
    def from_record_and_exception(
        cls, idx: int, record: "VCFRecord", e: Exception
    ) -> Self:
        """
        Create a VembraneError from a VCFRecord and an Exception.
        This is useful to preserve the record context in the error message.
        """
        return cls(
            f"Error processing record {idx}: {e}\nRecord: {record}",
        )

    def __str__(self) -> str:
        return self.args[0]


class UnknownAnnotationError(VembraneError):
    """Unknown annotation entry"""

    def __init__(self, record, key: str) -> None:
        super().__init__(
            f"No ANN entry for '{key}' in record {record.record_idx}:\n{str(record)}",
        )
        self.record_idx = record.record_idx
        self.record = record
        self.key = key


class MalformedAnnotationError(VembraneError):
    """Malformed annotation entry"""

    def __init__(self, record, key: str, ann_idx: int) -> None:
        super().__init__(
            f"The annotation index {ann_idx} ('{key}') "
            f"in record {record.record_idx} is out of bounds. "
            f"The ANN field might be malformed for this record:\n{str(record)}",
        )
        self.record_idx = record.record_idx
        self.record = record
        self.key = key
        self.ann_idx = ann_idx


class UnknownSampleError(VembraneError, KeyError):
    """Unknown Sample"""

    def __init__(self, record, sample: str) -> None:
        super().__init__(
            f"No sample with name '{sample}' in record \
                {record.record_idx}:\n{str(record)}",
        )
        self.record_idx = record.record_idx
        self.record = record
        self.field = sample


class UnknownFormatFieldError(VembraneError, KeyError):
    """Unknown FORMAT key"""

    def __init__(self, record, field: str) -> None:
        super().__init__(
            f"No FORMAT field '{field}' in record \
                {record.record_idx}:\n{str(record)}",
        )
        self.record_idx = record.record_idx
        self.record = record
        self.field = field


class UnknownInfoFieldError(VembraneError):
    """Unknown INFO key"""

    def __init__(self, record, field: str) -> None:
        super().__init__(
            f"No INFO field '{field}' in record {record.record_idx}:\n{str(record)}",
        )
        self.record_idx = record.record_idx
        self.record = record
        self.field = field


class InvalidExpressionError(VembraneError):
    """Filter expression is invalid"""

    def __init__(self, expression: str, reason: str) -> None:
        super().__init__(
            f"The provided expression '{expression}' is invalid. Reason: {reason}",
        )
        self.expression = expression


class MoreThanOneAltAlleleError(VembraneError):
    """vembrane only supports one ALT allele per record"""

    def __init__(self) -> None:
        msg = (
            "vembrane only supports records with one alternative allele.\n"
            "Please split multi-allelic records first, for example with "
            "`bcftools norm -m-any […]` or "
            "`gatk LeftAlignAndTrimVariants […] --split-multi-allelics` or "
            "`vcfmulti2oneallele […]`"
        )
        super().__init__(msg)


class NotExactlyOneValueError(VembraneError):
    """There may only be one value in VCF fields with `Number=1`"""

    def __init__(self, field: str, nvalues: int, record_idx: int) -> None:
        msg = (
            f"record {record_idx} has {nvalues} values in {field}, "
            f"but defined `Number=1`.\n"
            "vembrane enforces fields with `Number=1` to actually only have "
            "one number, "
            "(see VCF specification v4.3 section 1.4.2).\n"
            "To override this behaviour, "
            f"use `--overwrite-number-[info|format] {field} NUM`, "
            "where `NUM` is one of: \n"
            "- A: the field has one value per alternate allele\n"
            "- R: the field has one value for each possible allele, "
            "including the reference.\n"
            "- .: the number of possible values varies, is unknown or unbounded\n"
            "- G (FORMAT only): the field has one value for each possible genotype\n"
            "- {number}: the field has exactly {NUMBER} values\n"
        )
        super().__init__(msg)


class HeaderWrongColumnNumberError(VembraneError):
    """
    table --header expression generates different number of columns than main expression
    """

    def __init__(
        self,
        n_expr_cols: int,
        expr_cols: List[str],
        n_header_cols: int,
        header_cols: List[str],
    ) -> None:
        msg = (
            "The provided --header expression generates a different\n"
            "number of columns than the main expression:\n"
            f"{n_expr_cols} main expression columns: {expr_cols}\n"
            f"{n_header_cols} --header expression columns: {header_cols}\n"
        )
        super().__init__(msg)


class FilterAlreadyDefinedError(VembraneError):
    def __init__(self, tag: str) -> None:
        super().__init__(
            f"Filter {tag} already defined in the header. Choose a different name.",
        )


class FilterTagNameInvalidError(VembraneError):
    def __init__(self, tag: str) -> None:
        super().__init__(
            f"Filter '{tag}' contains invalid characters (whitespace or semicolon) "
            f"or is '0'.",
        )


class NonBoolTypeError(VembraneError):
    def __init__(self, value: Any):
        super().__init__(
            "The expression does not evaluate to bool, "
            f"but to {type(value)} ({value}).\n"
            "If you wish to use truthy values, "
            "explicitly wrap the expression in `bool(…)`, "
            "or aggregate multiple values via `any(…)` or `all(…)`.",
        )


class UnsupportedChromName(VembraneError):
    def __init__(self, chrom: str):
        super().__init__(f"Unsupported chromosome name: {chrom}.")


def handle_vembrane_error(func):
    """
    Decorator to handle VembraneError exceptions and print a user-friendly message.
    """

    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except VembraneError as e:
            print(f"Error: {e}", file=sys.stderr)
            sys.exit(1)

    return wrapper



import json
import sys
import textwrap
from typing import Any, Callable, Dict, Iterator, TextIO

import yaml
import yte  # type: ignore

from vembrane.ann_types import NA
from vembrane.backend.base import VCFHeader, VCFReader, VCFRecord
from vembrane.common import (
    Context,
    HumanReadableDefaultsFormatter,
    Primitive,
    add_common_arguments,
    create_reader,
    smart_open,
)
from vembrane.errors import VembraneError, handle_vembrane_error
from vembrane.representations import (
    Annotation,
    SourceEnvironment,
)


def add_subcommand(subparsers):
    parser = subparsers.add_parser(
        "structured",
        help="Create structured output (e.g., JSON/YAML) "
        "from a VCF/BCF file using a YTE template.",
        description="Create structured output from a VCF/BCF and a YTE template.",
        formatter_class=HumanReadableDefaultsFormatter,
    )
    parser.add_argument(
        "template",
        help="File containing a YTE template with the desired structure per record "
        "and expressions that retrieve data from the VCF/BCF record.",
    )
    parser.add_argument(
        "vcf",
        help="Path to the VCF/BCF file to be filtered. Defaults to '-' for stdin.",
        nargs="?",
        default="-",
    )
    parser.add_argument(
        "--output-fmt",
        choices=["json", "jsonl", "yaml"],
        help="Output format. If not specified, can be automatically determined from "
        "the --output file extension.",
    )
    parser.add_argument(
        "--output",
        "-o",
        default="-",
        help="Output file, if not specified, output is written to STDOUT.",
    )
    add_common_arguments(parser)


ConvertedRecords = Iterator[list[Any] | dict[Any, Any] | Any]


class CodeHandler(yte.CodeHandler):
    def __init__(
        self,
        ann_key: str,
        header: VCFHeader,
        allowed_globals: dict[str, Any] | None,
        auxiliary_globals: dict[str, Any] | None,
    ) -> None:
        self.ann_key = ann_key
        self.header = header
        self._envs: Dict[str, SourceEnvironment] = {}
        self._record: VCFRecord | None = None
        self._record_idx: int | None = None
        self._annotation: str | None = None
        self._auxiliary_globals = auxiliary_globals
        self._allowed_globals = allowed_globals

    def update_from_record(self, idx: int, record: VCFRecord) -> None:
        self._record = record
        self._record_idx = idx
        for env in self._envs.values():
            env.update_from_record(idx, record)

    def update_from_annotation(self, annotation: str | None) -> None:
        self._annotation = annotation
        if annotation is not None:
            for env in self._envs.values():
                if env.expression_annotations():
                    env.update_annotation(annotation)

    def _env(self, source: str) -> SourceEnvironment:
        env = self._envs.get(source)
        if env is None:
            env = SourceEnvironment(
                source,
                self.ann_key,
                self.header,
                allowed_globals=self._allowed_globals,
                auxiliary_globals=self._auxiliary_globals,
            )
            env.update_from_record(self._record_idx, self._record)  # type: ignore
            if self._annotation is not None:
                env.update_annotation(self._annotation)
            self._envs[source] = env
        return env

    def eval(self, expr: str, variables: Dict[str, Any]) -> Any:
        env = self._env(expr)
        return eval(env.compiled, variables, env)

    def exec(self, source: str, variables: Dict[str, Any]) -> Any:
        env = self._env(source)
        return exec(env.compiled, variables, env)

    def has_annotation(self) -> bool:
        return any(env.expression_annotations() for env in self._envs.values())


class ValueHandler(yte.ValueHandler):
    def postprocess_atomic_value(self, value: Any) -> Any:
        processed = super().postprocess_atomic_value(value)
        if processed is NA:
            return None
        return processed


def process_vcf(
    vcf: VCFReader,
    template: str,
    ann_key: str,
    allowed_globals: dict[str, Any] | None = None,
    auxiliary_globals: dict[str, Any] | None = None,
    postprocess: Callable[[Any], Any] | None = None,
) -> ConvertedRecords:
    code_handler = CodeHandler(ann_key, vcf.header, allowed_globals, auxiliary_globals)
    value_handler = ValueHandler()

    annotation = Annotation(ann_key, vcf.header)

    for idx, record in enumerate(vcf):
        try:
            # TODO: For now structural variants are not supported
            # but should be added later.
            if record.is_sv_record:
                print(
                    f"Warning: Record is a structural variant which are currently "
                    f"not supported and will be skipped.\n"
                    f"Record: {record}"
                )
                continue
            code_handler.update_from_record(idx, record)

            annotations: list[str] | None

            if annotation:
                annotations = annotation.get_record_annotations(idx, record)
            else:
                annotations = None

            # type ignore below because mypy is too stupid!
            for ann_entry in annotations or [None]:  # type: ignore
                if ann_entry is not None:
                    code_handler.update_from_annotation(ann_entry)
                converted = yte.process_yaml(
                    template,
                    code_handler=code_handler,
                    value_handler=value_handler,
                )
                if postprocess is not None:
                    converted = postprocess(converted)
                yield converted
                if not code_handler.has_annotation():
                    # If annotation is not used in template, only process first entry.
                    # There can be no subsequent entries leading to different results.
                    break

        except Exception as e:
            raise VembraneError.from_record_and_exception(idx, record, e) from e


def write_records_jsonl(output_file: TextIO, records: ConvertedRecords) -> None:
    for record in records:
        print(json.dumps(record), file=output_file)


def write_records_json(output_file: TextIO, records: ConvertedRecords) -> None:
    print("[", file=output_file)
    for idx, record in enumerate(records):
        if idx > 0:
            print(",", file=output_file)
        print(json.dumps(record), file=output_file)
    print("]", file=output_file)


def write_records_yaml(output_file: TextIO, records: ConvertedRecords) -> None:
    for record in records:
        head, tail = yaml.dump(record).split("\n", 1)
        head = f"- {head}"
        tail = textwrap.indent(tail, "  ")

        print(head, file=output_file)
        print(tail, file=output_file)


@handle_vembrane_error
def process(
    template: str,
    output_fmt: str | None,
    output: str,
    vcf: str,
    annotation_key: str,
    overwrite_number_info: dict[str, str],
    overwrite_number_format,
    backend,
    allowed_globals: Dict[str, Any] | None = None,
    auxiliary_globals: Dict[str, Any] | None = None,
    postprocess: (
        Callable[[Primitive | dict | list], Primitive | dict | list] | None
    ) = None,
) -> None:
    overwrite_number = {
        "INFO": dict(overwrite_number_info),
        "FORMAT": dict(overwrite_number_format),
    }

    if output_fmt is None:
        if output.endswith(".json"):
            output_fmt = "json"
        elif output.endswith(".jsonl"):
            output_fmt = "jsonl"
        elif output.endswith(".yaml") or output.endswith(".yml"):
            output_fmt = "yaml"
        elif output == "-":
            raise VembraneError(
                "--output-fmt must be specified when writing to STDOUT."
            )
        else:
            raise VembraneError(
                f"Unsupported file format: {output}, only .json, .jsonl and "
                ".yaml are supported."
            )

    if output_fmt == "jsonl":
        write_records = write_records_jsonl
    elif output_fmt == "json":
        write_records = write_records_json
    elif output_fmt == "yaml":
        write_records = write_records_yaml
    else:
        raise VembraneError(
            f"Unsupported output format: {output_fmt}, only json, jsonl and yaml are "
            "supported."
        )

    with (
        create_reader(
            vcf,
            backend=backend,
            overwrite_number=overwrite_number,
        ) as reader,
        smart_open(output, mode="w") as writer,
    ):
        write_records(
            writer,
            process_vcf(
                reader,
                template,
                annotation_key,
                allowed_globals,
                auxiliary_globals,
                postprocess=postprocess,
            ),
        )


def execute(args):
    try:
        with open(args.template, "r") as template:
            template = template.read()

        process(
            template=template,
            output_fmt=args.output_fmt,
            output=args.output,
            vcf=args.vcf,
            annotation_key=args.annotation_key,
            overwrite_number_info=args.overwrite_number_info,
            overwrite_number_format=args.overwrite_number_format,
            backend=args.backend,
            auxiliary_globals=Context.from_args(args).get_globals(),
        )
    except VembraneError as ve:
        print(ve, file=sys.stderr)
        sys.exit(1)



import argparse
import sys
from collections import defaultdict
from collections.abc import Iterator
from itertools import chain, islice
from sys import stderr
from typing import Any

import yaml

from .. import __version__
from ..backend.base import VCFReader, VCFRecord
from ..common import (
    BreakendEvent,
    Context,
    HumanReadableDefaultsFormatter,
    add_common_arguments,
    check_expression,
    create_reader,
    create_writer,
    get_annotation_keys,
    mate_key,
    normalize,
    read_auxiliary,
    read_ontology,
    split_annotation_entry,
)
from ..errors import VembraneError, handle_vembrane_error
from ..representations import FuncWrappedExpressionEnvironment
from ..sequence_ontology import SequenceOntology


class DeprecatedAction(argparse.Action):
    def __call__(self, *args, **kwargs):
        print(
            f"{'/'.join(self.option_strings)} is deprecated.\n{self.help}",
            file=sys.stderr,
        )
        sys.exit(1)


def add_subcommmand(subparsers):
    parser = subparsers.add_parser(
        "filter",
        help="Filter VCF/BCF records and annotations using a python expression.",
        description="Filter VCF/BCF records and annotations "
        "based on a user-defined Python expression."
        "Only records for which the expression evaluates to True are kept.",
        formatter_class=HumanReadableDefaultsFormatter,
    )
    parser.register("action", "deprecated", DeprecatedAction)
    parser.add_argument(
        "expression",
        type=check_expression,
        help="A python expression to filter variants. "
        "The expression must evaluate to bool. "
        "All VCF/BCF fields can be accessed. "
        "Additionally, annotation fields can be accessed, see `--annotation-key`. "
        "If all annotations of a record are filtered out, "
        "the entire record is removed.",
    )
    parser.add_argument(
        "vcf",
        help="Path to the VCF/BCF file to be filtered. Defaults to '-' for stdin.",
        nargs="?",
        default="-",
    )
    parser.add_argument(
        "--output",
        "-o",
        default="-",
        help="Output file, if not specified, output is written to STDOUT.",
    )
    parser.add_argument(
        "--output-fmt",
        "-O",
        default="vcf",
        choices=["vcf", "bcf", "uncompressed-bcf"],
        help="Output format.",
    )
    parser.add_argument(
        "--statistics",
        "-s",
        metavar="FILE",
        default=None,
        help="Write statistics to this file.",
    )
    parser.add_argument(
        "--keep-unmatched",
        default=False,
        action="store_true",
        help="Keep all annotations of a variant if at least one of them passes "
        "the expression.",
    )
    parser.add_argument(
        "--preserve-order",
        default=False,
        action="store_true",
        help="Ensures that the order of the output matches that of the \
              input. This is only useful if the input contains breakends (BNDs) \
              since the order of all other variants is preserved anyway.",
    )
    add_common_arguments(parser)


def test_and_update_record(
    env: FuncWrappedExpressionEnvironment,
    idx: int,
    record: VCFRecord,
    ann_key: str,
    keep_unmatched: bool,
) -> tuple[VCFRecord, bool]:
    try:
        return _test_and_update_record(env, idx, record, ann_key, keep_unmatched)
    except VembraneError as e:
        raise e
    except Exception as e:
        raise VembraneError.from_record_and_exception(idx, record, e) from e


def _test_and_update_record(
    env: FuncWrappedExpressionEnvironment,
    idx: int,
    record: VCFRecord,
    ann_key: str,
    keep_unmatched: bool,
) -> tuple[VCFRecord, bool]:
    env.update_from_record(idx, record)
    if env.expression_annotations():
        annotations = env.get_record_annotations(idx, record)

        #  … and only keep the annotations where the expression evaluates to true
        if keep_unmatched:
            filtered = any(map(env.is_true, annotations))
            return record, filtered
        else:
            filtered_annotations = [
                annotation for annotation in annotations if env.is_true(annotation)
            ]

            if len(annotations) != len(filtered_annotations):
                # update annotations if they have actually been filtered
                record.info[ann_key] = filtered_annotations

            return record, len(filtered_annotations) > 0
    else:
        # otherwise, the annotations are irrelevant w.r.t. the expression,
        # so we can omit them
        return record, env.is_true()


def filter_vcf(
    reader: VCFReader,
    expression: str,
    ann_key: str,
    keep_unmatched: bool = False,
    preserve_order: bool = False,
    auxiliary: dict[str, set[str]] | None = None,
    auxiliary_globals: dict[str, Any] | None = None,
    ontology: SequenceOntology | None = None,
) -> Iterator[VCFRecord]:
    if auxiliary is None:
        auxiliary = {}
    env = FuncWrappedExpressionEnvironment(
        expression,
        ann_key,
        reader.header,
        auxiliary,
        ontology,
        auxiliary_globals=auxiliary_globals,
    )
    has_mateid_key = reader.header.infos.get("MATEID", None) is not None
    has_event_key = reader.header.infos.get("EVENT", None) is not None

    def get_event_name(
        record: VCFRecord,
        has_mateid_key: bool = has_mateid_key,
        has_event_key: bool = has_event_key,
    ) -> tuple[str | None, str | None]:
        mate_ids: list[str] | str = (
            record.info.get("MATEID", []) if has_mateid_key else []
        )
        if isinstance(mate_ids, str):
            # some callers annotate MATEID as single string
            mate_ids = [mate_ids]
        event_name: str | None = (
            record.info.get("EVENT", None) if has_event_key else None
        )

        if len(mate_ids) > 1 and not event_name:
            raise ValueError(
                f"Filtering of BND records with multiple mates is unsupported "
                f"(see VCF 4.3, section 5.4.3 'Multiple mates'):\n{str(record)}",
            )

        mate_id = mate_ids[0] if len(mate_ids) == 1 else None
        mate_pair = mate_key([record.id, mate_id]) if mate_ids else None

        return event_name, mate_pair

    record: VCFRecord
    if not preserve_order:
        # If the order is not important, emit records that pass the filter expression
        # as we encounter them
        # However, breakends have to be considered jointly, so keep track of the
        # respective events.
        event_dict: dict[str, BreakendEvent] = {}
        for idx, record in enumerate(reader):
            record, keep = test_and_update_record(
                env,
                idx,
                record,
                ann_key,
                keep_unmatched,
            )

            # Breakend records *may* have the "EVENT" specified, but don't have to.
            # In that case only the MATEID *may* be available
            # (which may contain more than one ID)
            if record.is_bnd_record:
                event_name, mate_pair_name = get_event_name(record)

                # if EVENT is set, it has priority over MATEID.
                event_name = event_name or mate_pair_name

                # if both EVENT and MATEID are not set
                # we can only treat this BND record as a regular one,
                # (or somehow try to guess its correct mate)
                if not event_name:
                    print(
                        f"Warning: Encountered breakend record at index {idx} "
                        f"without either of MATEID or EVENT specified, "
                        f"treating it as a regular record:\n{str(record)}",
                        file=stderr,
                    )
                    if keep:
                        yield record
                    continue

                event = event_dict.get(event_name, None)

                # if there's already an associated event
                if event:
                    # add this record to the event
                    event.add(record, keep)

                    # if we already know that the event is a "PASS"…
                    if event.keep:
                        # … emit all records associated with it
                        yield from event.emit()

                        # in the case of a simple mate pair, we can delete the event
                        # at this point, because no more records will be added to it
                        if event.is_mate_pair() and mate_pair_name:
                            del event_dict[mate_pair_name]
                else:
                    # if there's no entry for the event or mate pair yet, create one
                    is_mate_pair = bool(mate_pair_name) and mate_pair_name == event_name
                    event = BreakendEvent(event_name, is_mate_pair)
                    event.add(record, keep)
                    event_dict[event_name] = event
            elif keep:
                yield record

        if len(event_dict) > 0:
            # output BNDs if any are left unprocessed
            for event_name, event in event_dict.items():
                if event_name and event.keep:
                    yield from event.emit()
    elif preserve_order:
        # If order *is* important, the first pass cannot emit any records but only
        # keep track of breakend events. The records will only be emitted during the
        # second pass.
        event_set: set[str] = set()

        def fallback_name(record: VCFRecord) -> str:
            event_name, mate_pair_name = get_event_name(record)

            # There may be BND record which have neither of
            # INFO['EVENT'], INFO['MATEID'] or even ID specified.
            # In that case, we still need a pseudo-event name,
            # which we construct from the record's index in the file.
            return event_name or mate_pair_name or record.id or f"DUMMY: {idx}"

        for idx, record in enumerate(reader):
            if record.is_bnd_record:
                record, keep = test_and_update_record(
                    env,
                    idx,
                    record,
                    ann_key,
                    keep_unmatched,
                )
                if keep:
                    event_name = fallback_name(record)
                    event_set.add(event_name)

        # The second pass can now yield records in the correct order
        reader.reset()
        for idx, record in enumerate(reader):
            if record.is_bnd_record:
                event_name = fallback_name(record)
                if event_name in event_set:
                    yield record
            else:
                record, keep = test_and_update_record(
                    env,
                    idx,
                    record,
                    ann_key,
                    keep_unmatched,
                )
                if keep:
                    yield record


def statistics(
    records: Iterator[VCFRecord],
    vcf: VCFReader,
    filename: str,
    ann_key: str,
) -> Iterator[VCFRecord]:
    annotation_keys = get_annotation_keys(vcf.header, ann_key)
    counter: defaultdict[str, defaultdict[str, Any]] = defaultdict(
        lambda: defaultdict(lambda: 0)
    )
    for record in records:
        for annotation in record.info[ann_key]:
            for key, raw_value in zip(
                annotation_keys,
                split_annotation_entry(annotation),
                strict=True,
            ):
                value = raw_value.strip()
                if value:
                    counter[key][value] += 1
        yield record

    # reduce dicts with many items, to just one counter
    for key, subdict in counter.items():
        if len(subdict) > 10:
            counter[key] = f"#{len(subdict)}"  # type: ignore

    yaml.add_representer(defaultdict, yaml.representer.Representer.represent_dict)
    with open(filename, "w") as out:
        yaml.dump(dict(counter), out)


@handle_vembrane_error
def execute(args) -> None:
    aux = read_auxiliary(args.aux)
    ontology = read_ontology(args.ontology)

    overwrite_number = {
        "INFO": dict(args.overwrite_number_info),
        "FORMAT": dict(args.overwrite_number_format),
    }
    with create_reader(
        args.vcf,
        backend=args.backend,
        overwrite_number=overwrite_number,
    ) as reader:
        reader.header.add_generic("vembraneVersion", __version__)
        # NOTE: If .modules.filter.execute might be used as a library function
        #       in the future, we should not record sys.argv directly below.
        cmd_parts = [normalize(arg) if " " in arg else arg for arg in sys.argv[3:]]
        expr = " ".join(a.strip() for a in args.expression.split("\n"))
        expr = normalize(expr)

        reader.header.add_generic(
            "vembraneCmd",
            "vembrane " + expr + " " + " ".join(cmd_parts),
        )

        records = filter_vcf(
            reader,
            args.expression,
            args.annotation_key,
            keep_unmatched=args.keep_unmatched,
            preserve_order=args.preserve_order,
            auxiliary=aux,
            auxiliary_globals=Context.from_args(args).get_globals(),
            ontology=ontology,
        )

        first_record = list(islice(records, 1))

        records = chain(first_record, records)

        if args.statistics is not None:
            records = statistics(records, reader, args.statistics, args.annotation_key)

        fmt = {"vcf": "", "bcf": "b", "uncompressed-bcf": "u"}[args.output_fmt]

        with create_writer(args.output, fmt, reader, backend=args.backend) as writer:
            for record in records:
                writer.write(record)



from collections.abc import Callable, Iterable, Iterator
from sys import exit, stderr
from typing import Any

import numpy as np
import yaml
from intervaltree import Interval, IntervalTree  # type: ignore

from vembrane.errors import handle_vembrane_error  # type: ignore

from ..backend.base import VCFReader, VCFRecord
from ..common import (
    Context,
    HumanReadableDefaultsFormatter,
    add_common_arguments,
    check_expression,
    create_reader,
    create_writer,
)
from ..representations import FuncWrappedExpressionEnvironment


def add_subcommmand(subparsers):
    parser = subparsers.add_parser(
        "annotate",
        help="Add new INFO field annotations to a VCF/BCF from other data sources, "
        "using a configuration file.",
        description="Add new INFO field annotations to a VCF/BCF from other data "
        "sources, using a configuration file.",
        formatter_class=HumanReadableDefaultsFormatter,
    )
    parser.add_argument(
        "config",
        type=check_expression,
        help="The configuration file.",
    )
    parser.add_argument(
        "vcf",
        help="Path to the VCF/BCF file to be filtered. Defaults to '-' for stdin.",
        nargs="?",
        default="-",
    )
    parser.add_argument(
        "--output",
        "-o",
        default="-",
        help="Output file, if not specified, output is written to STDOUT.",
    )
    parser.add_argument(
        "--output-fmt",
        "-O",
        default="vcf",
        choices=["vcf", "bcf", "uncompressed-bcf"],
        help="Output format.",
    )
    add_common_arguments(parser)


typeparser: dict[str, Callable[[str], Any]] = {
    "Float": float,
    "Integer": int,
    "Character": lambda x: str(x)[0],
    "String": str,
    "Flag": bool,
}


def annotate_vcf(
    vcf: VCFReader,
    expression: str,
    ann_key: str,
    ann_data,
    config: dict,
    auxiliary_globals: dict[str, Any] | None = None,
) -> Iterator[VCFRecord]:
    env = FuncWrappedExpressionEnvironment(
        expression, ann_key, vcf.header, auxiliary_globals=auxiliary_globals
    )

    config_chrom_column: str = config["annotation"]["columns"]["chrom"]
    config_start_column: str = config["annotation"]["columns"]["start"]
    config_stop_column: str = config["annotation"]["columns"]["stop"]

    available_chromsomes: set[str] = set(np.unique(ann_data[config_chrom_column]))

    tree: dict[str, IntervalTree] = {}
    chrom_ann_data: dict[str, Any] = {}
    for available_chrom in available_chromsomes:
        d = ann_data[ann_data[config_chrom_column] == available_chrom]
        chrom_ann_data[available_chrom] = d
        tree[available_chrom] = IntervalTree(
            Interval(d[config_start_column], d[config_stop_column], i)
            for i, d in enumerate(d)
        )

    current_chrom = None
    current_ann_data = None

    record: VCFRecord
    for idx, record in enumerate(vcf):
        chrom: str | None = None
        if current_chrom != record.contig:
            current_chrom = record.contig

            # find the correct chrom name
            tmp = current_chrom
            if tmp.lower().startswith("chr"):
                tmp = tmp[3:]
            for prefix in ["", "chr", "Chr", "CHR"]:
                if prefix + tmp in available_chromsomes:
                    chrom = prefix + tmp
                    current_ann_data = chrom_ann_data[chrom]
                    t = tree[chrom]

        if chrom:
            indices = np.fromiter((i for _, _, i in t[record.position]), dtype=int)
            if len(indices):
                env.update_data(current_ann_data[(np.array(indices))])  # type: ignore
                env.update_from_record(idx, record)
                ann_values = env.table_row()

                for v, expression_value in zip(
                    (x["value"] for x in config["annotation"]["values"]),
                    ann_values,
                    strict=True,
                ):
                    number = int(v["number"]) if v["number"] != "." else -1

                    parse = typeparser[v["type"]]
                    if number == -1:
                        ev = list(map(parse, expression_value))
                    elif number > 1:
                        ev = list(map(parse, expression_value))
                        assert len(ev) == number
                    else:
                        # number == 1
                        assert isinstance(expression_value, str) or not isinstance(
                            expression_value,
                            Iterable,
                        )
                        ev = parse(expression_value)
                    record.info[v["vcf_name"]] = ev

        yield record


@handle_vembrane_error
def execute(args):
    with open(args.config) as stream:
        try:
            config = yaml.safe_load(stream)
        except yaml.YAMLError as e:
            print(e, file=stderr)
            exit(1)

    # load annotation data
    ann_data: np.ndarray = np.genfromtxt(
        config["annotation"]["file"],
        delimiter=config["annotation"].get("delimiter", "\t"),
        names=True,
        dtype=None,
        encoding=None,
    )

    # build expression
    expression = ",".join(
        f"{value['expression']}"
        for value in (x["value"] for x in config["annotation"]["values"])
    )
    expression = f"({expression})"

    with create_reader(args.vcf) as vcf:
        # add new info
        for value in config["annotation"]["values"]:
            v = value["value"]
            vcf.header.add_meta(
                "INFO",
                items=[
                    ("ID", v["vcf_name"]),
                    ("Number", v["number"]),
                    ("Type", v["type"]),
                    ("Description", v["description"]),
                ],
            )

        fmt = {"vcf": "", "bcf": "b", "uncompressed-bcf": "u"}[args.output_fmt]
        with create_writer(
            args.output,
            f"w{fmt}",
            header=vcf.header,
        ) as out:
            variants = annotate_vcf(
                vcf,
                expression,
                args.annotation_key,
                ann_data=ann_data,
                config=config,
                auxiliary_globals=Context.from_args(args).get_globals(),
            )
            for v in variants:
                out.write(v)



from collections import OrderedDict, defaultdict
from pathlib import Path
from sys import stderr
from typing import Dict, List, Optional, Tuple

import numpy as np
from cyvcf2.cyvcf2 import VCF, Variant, Writer  # type: ignore

from vembrane.backend.base import (
    VCFHeader,
    VCFReader,
    VCFRecord,
    VCFRecordFilter,
    VCFRecordFormat,
    VCFRecordFormats,
    VCFRecordInfo,
    VCFWriter,
)

from ..ann_types import NA, type_info
from ..errors import UnknownInfoFieldError, UnknownSampleError


class Cyvcf2Reader(VCFReader):
    __slots__ = (
        "filename",
        "_iter_file",
        "_header",
        "_overwrite_number",
        "_current_record_idx",
    )

    def __init__(
        self,
        filename: str | Path,
        overwrite_number: Dict[str, Dict[str, str]] | None = None,
    ):
        if overwrite_number is None:
            overwrite_number = {}
        self.filename = filename
        self._file = VCF(self.filename)
        self._header = Cyvcf2Header(self, overwrite_number)
        self._overwrite_number = overwrite_number
        self._current_record_idx = 0

    def __iter__(self):
        self._iter_file = self._file.__iter__()
        return self

    def __next__(self):
        self._current_record_idx += 1
        return Cyvcf2Record(
            self._iter_file.__next__(),
            self._current_record_idx,
            self._header,
            self._file,
        )

    def reset(self):
        # cyvcv2 doesnt have a reset function
        metadata_generic = self._header._metadata_generic.copy()
        self._file.close()
        self._file = VCF(self.filename)
        self._header = Cyvcf2Header(self, self._overwrite_number)
        for k, v in metadata_generic.items():
            self._header.add_generic(k, v)
        self._current_record_idx = 0
        # TODO: may this workaround lead to problems?


class Cyvcf2Header(VCFHeader):
    __slots__ = ("_reader", "_data", "_data_category", "_metadata_generic")

    def __init__(
        self,
        reader: Cyvcf2Reader,
        overwrite_number: Dict[str, Dict[str, str]],
    ):
        self._reader = reader
        self._data = []
        self._data_category: defaultdict[str | None, OrderedDict] = defaultdict(
            OrderedDict
        )
        self._metadata_generic = dict()

        for r in reader._file.header_iter():
            if r.type == "GENERIC":
                continue
            d = r.info()
            self._data.append(d)
            if "ID" in d:
                self._data_category[r.type][d["ID"]] = d

        specific_keys = {r.type for r in reader._file.header_iter()} | {"contig"}
        _generic_entries = [
            r.lstrip("#").split("=", 1)
            for r in reader._file.raw_header.split("\n")
            if r.startswith("##")
        ]
        generic_entries: dict[str, str] = {
            k: v for (k, v) in _generic_entries if k not in specific_keys
        }
        for k, v in generic_entries.items():
            self._metadata_generic[k] = v

        # override numbers
        for category, items in overwrite_number.items():
            for key, value in items.items():
                if key in self._data_category[category]:
                    self._data_category[category][key]["Number"] = value

    @property
    def records(self):
        return self._data_category

    def contains_generic(self, key: str):
        return key in self._metadata_generic

    def get_generic(self, key: str):
        return self._metadata_generic[key]

    @property
    def infos(self):
        return self._data_category["INFO"]

    @property
    def formats(self):
        return self._data_category["FORMAT"]

    @property
    def filters(self):
        return self._data_category["FILTER"]

    @property
    def samples(self):
        return self._reader._file.samples

    def add_generic(
        self,
        key: str,
        value: str,
    ):
        self._metadata_generic[key] = value
        self._reader._file.add_to_header(f"##{key}={value}")

    def add_filter(self, id: str, description: str):
        self._reader._file.add_filter_to_header({"ID": id, "Description": description})

    def __iter__(self):
        raise NotImplementedError

    def __next__(self):
        raise NotImplementedError

    def add_meta(
        self,
        key: str,
        value: Optional[str] = None,
        items: Optional[List[Tuple[str, str]]] = None,
    ):
        raise NotImplementedError


class Cyvcf2Record(VCFRecord):
    __slots__ = ("_file",)

    def __init__(
        self,
        record: Variant,
        record_idx: int,
        header: Cyvcf2Header,
        file: VCF,
    ):
        super().__init__(record, record_idx, header)
        self._file = file

    @property
    def contig(self) -> str:
        return self._raw_record.CHROM

    @property
    def position(self) -> int:
        return self._raw_record.POS

    @property
    def stop(self) -> int:
        if "END" in self.info:
            return self.info["END"]
        return self._raw_record.POS + len(self._raw_record.REF) - 1

    @property
    def id(self) -> str:
        return self._raw_record.ID

    @property
    def reference_allele(self) -> str:
        return self._raw_record.REF

    @property
    def alt_alleles(self) -> Tuple[str]:
        return tuple(self._raw_record.ALT)

    @property
    def quality(self) -> float:
        return self._raw_record.QUAL

    @property
    def filter(self) -> VCFRecordFilter:
        return Cyvcf2RecordFilter(self)

    @property
    def info(self) -> VCFRecordInfo:
        return Cyvcf2RecordInfo(self)

    @property
    def formats(self) -> VCFRecordFormats:
        return Cyvcf2RecordFormats(self)

    @property
    def samples(self):
        return self._file.samples

    @property
    def header(self) -> VCFHeader:
        return self._header


class Cyvcf2RecordFormats(VCFRecordFormats):
    __slots__ = ("_record", "_raw_format", "_format")

    def __init__(self, record: Cyvcf2Record):
        self._record = record
        self._raw_format = record._raw_record.FORMAT
        self._format: dict[str, Cyvcf2RecordFormat] = {}

    def __getitem__(self, key: str):
        if key not in self._format:
            self._format[key] = Cyvcf2RecordFormat(key, self._record)
        return self._format[key]

    def __contains__(self, key):
        return key in self._raw_format

    def keys(self):
        return self._raw_format.keys()


class Cyvcf2RecordFormat(VCFRecordFormat):
    __slots__ = ("_header", "_record", "_raw_record", "_format_key")

    def __init__(self, format_key: str, record: Cyvcf2Record):
        self._record = record
        self._raw_record = record._raw_record
        self._header = record._header
        self._format_key = format_key

    def __getitem__(self, sample):
        i = self._header.samples.index(sample)
        if i == -1:
            raise UnknownSampleError(self._record, sample)
        if self._format_key == "GT":  # genotype
            value = tuple(
                None if gt == -1 else gt for gt in self._raw_record.genotypes[i][:-1]
            )
            return type_info(value, ".")

        meta = self._header.formats[self._format_key]
        number = meta["Number"]
        value_array = self._raw_record.format(self._format_key)
        if value_array is None:
            print(
                f"Warning: "
                f"record {self._record.record_idx} is missing a value "
                f"for FORMAT key {self._format_key}, "
                f"returning NA instead."
                f"\n{self._record}\n",
                file=stderr,
            )
            return type_info(NA, number)

        else:
            value = value_array[i]

        if meta["Type"] == "String" and not number == "1":
            value = value.split(",")
        if number == "1":
            if isinstance(value, np.ndarray):
                value = value[0].item()
            # cyvcf2 gives min int for unknown integer values
            if meta["Type"] == "Integer" and value == np.iinfo(np.int32).min:
                return NA
        return type_info(value, number)

    def __setitem__(self, key, value):
        raise NotImplementedError

    def __contains__(self, sample):
        return sample in self._header.samples

    def __eq__(self, other):
        return all(self[sample] == other[sample] for sample in self._header.samples)

    def keys(self):
        return self._raw_record.format.keys()


class Cyvcf2RecordFilter(VCFRecordFilter):
    __slots__ = "_record"

    def __init__(self, record: Cyvcf2Record):
        self._record = record._raw_record

    def __iter__(self):
        yield from self._record.FILTERS

    def add(self, tag: str):
        # cyvcf needs a semicolon separated string
        filter = self._record.FILTER
        if filter:
            filter += f";{tag}"
        else:
            filter = tag
        self._record.FILTER = filter

    def __contains__(self, key):
        return key in self._record.FILTERS


class Cyvcf2RecordInfo(VCFRecordInfo):
    __slots__ = ("_record", "_raw_record", "_header")

    def __init__(
        self,
        record: Cyvcf2Record,
    ):
        self._record = record
        self._raw_record = record._raw_record
        self._header = record._header

    def __getitem__(self, key):
        try:
            meta = self._header.infos[key]
            number = meta["Number"]
        except KeyError as ke:
            raise UnknownInfoFieldError(self._record, key) from ke

        try:
            value = self._raw_record.INFO[key]
            typ = meta["Type"]
        except KeyError:
            if key == "END":
                return self._raw_record.POS + len(self._raw_record.REF) - 1

            print(
                f"Warning: "
                f"record {self._record.record_idx} is missing a value "
                f"for INFO key {key}, "
                f"returning NA instead."
                f"\n{self._record}\n",
                file=stderr,
            )
            return type_info(NA, number)

        # for some reason cyvcf2 doesn't split String lists, a known circumstance
        if typ == "String" and not number == "1":
            value = value.split(",")

        return type_info(value, number)

    def __setitem__(self, key, value):
        # for some reason cyvcf2 doesn't split String lists, a known circumstance
        meta = self._header.infos[key]
        number, typ = meta["Number"], meta["Type"]
        if typ == "String" and number != "1":
            value = ",".join(value)
        self._raw_record.INFO[key] = value

    def __contains__(self, key):
        return self._raw_record.INFO.get(key, None) is not None

    def keys(self):
        return self._raw_record.info.keys()


class Cyvcf2Writer(VCFWriter):
    def __init__(self, filename: str | Path, fmt: str, template: VCFReader):
        self._file = Writer(filename, template.file, mode=f"w{fmt}")

    def write(self, record: VCFRecord):
        self._file.write_record(record._raw_record)



import csv
from collections.abc import Iterator
from enum import Enum
from typing import Any

import asttokens

from ..backend.base import VCFHeader, VCFReader, VCFRecord
from ..common import (
    Context,
    HumanReadableDefaultsFormatter,
    add_common_arguments,
    check_expression,
    create_reader,
    get_annotation_keys,
    read_auxiliary,
    read_ontology,
    smart_open,
)
from ..errors import HeaderWrongColumnNumberError, handle_vembrane_error
from ..globals import default_allowed_globals
from ..representations import FuncWrappedExpressionEnvironment
from ..sequence_ontology import SequenceOntology
from .filter import DeprecatedAction

ALL_EXPRESSION = "ALL"


def add_subcommmand(subparsers):
    parser = subparsers.add_parser(
        "table",
        help="Convert VCF/BCF records to tabular format.",
        description="Convert VCF/BCF records to tabular format.",
        formatter_class=HumanReadableDefaultsFormatter,
    )
    parser.register("action", "deprecated", DeprecatedAction)
    parser.add_argument(
        "expression",
        type=check_expression,
        help="A comma-separated tuple of expressions "
        "that define the table column contents. "
        f"Use {ALL_EXPRESSION} to output all fields.",
    )
    parser.add_argument(
        "vcf",
        help="Path to the VCF/BCF file to be filtered. Defaults to '-' for stdin.",
        nargs="?",
        default="-",
    )
    parser.add_argument(
        "--separator",
        "-s",
        default="\t",
        metavar="CHAR",
        help="Define the field separator (default: \\t).",
    )
    parser.add_argument(
        "--header",
        default="auto",
        metavar="TEXT",
        help='Override the automatically generated header. Provide "auto" (default) \
              to automatically generate the header from the expression. Provide a \
              comma separated string to manually set the header. Provide "none" to \
              disable any header output.',
    )
    parser.add_argument(
        "--naming-convention",
        metavar="CONVENTION",
        default="dictionary",
        type=NamingConvention,
        choices=list(NamingConvention),
        help="The naming convention to use for column names "
        "when generating the header for the ALL expression.",
    )
    parser.add_argument(
        "--wide",
        help="Instead of using long format with a special SAMPLE column, "
        "generate multiple columns per sample "
        "with the `for_each_sample` utility function.",
        action="store_true",
    )
    parser.add_argument(
        "--long",
        action=DeprecatedAction,
        help="Long format is now the default. For wide format, use `--wide` instead.",
    )
    parser.add_argument(
        "--output",
        "-o",
        default="-",
        help="Output file, if not specified, output is written to STDOUT.",
    )
    add_common_arguments(parser)


def tableize_vcf(
    vcf: VCFReader,
    expression: str,
    ann_key: str,
    overwrite_number: dict[str, dict[str, str]] | None = None,
    wide: bool = False,
    auxiliary: dict[str, set[str]] | None = None,
    auxiliary_globals: dict[str, Any] | None = None,
    ontology: SequenceOntology | None = None,
) -> Iterator[tuple]:
    if overwrite_number is None:
        overwrite_number = {}
    if auxiliary is None:
        auxiliary = {}

    kwargs: dict[str, Any] = dict(
        auxiliary_globals=auxiliary_globals, auxiliary=auxiliary, ontology=ontology
    )

    long_with_samples = not wide and list(vcf.header.samples)
    if long_with_samples:
        kwargs["evaluation_function_template"] = (
            "lambda: (({expression}) for SAMPLE in SAMPLES)"
        )
    else:
        expression = f"({expression})"
    env = FuncWrappedExpressionEnvironment(expression, ann_key, vcf.header, **kwargs)

    record: VCFRecord
    for idx, record in enumerate(vcf):
        env.update_from_record(idx, record)
        if env.expression_annotations():
            annotations = env.get_record_annotations(idx, record)
            for annotation in annotations:
                if long_with_samples:
                    yield from env.table_row(annotation)
                else:
                    yield env.table_row(annotation)
        else:
            if long_with_samples:
                yield from env.table_row()
            else:
                yield env.table_row()


class NamingConvention(Enum):
    DICTIONARY = "dictionary"
    UNDERSCORE = "underscore"
    SLASH = "slash"


def construct_all_expression_and_header(
    header: VCFHeader, ann_key: str, naming_convention: NamingConvention
) -> tuple[str, list[str]]:
    all_header = ["SAMPLE", "CHROM", "POS", "REF", "ALT", "QUAL", "FILTER", "ID"]
    all_expression = ", ".join(all_header)

    def format_key(field: str, key: str) -> str:
        match naming_convention:
            case NamingConvention.DICTIONARY:
                return f"{field}['{key}']"
            case NamingConvention.UNDERSCORE:
                return f"{field}_{key}"
            case NamingConvention.SLASH:
                return f"{field}/{key}"

    info_keys = list(header.infos.keys())
    if info_keys:
        info_keys_without_ann = [key for key in info_keys if key != ann_key]
        info_expr = ", ".join(f'INFO["{key}"]' for key in info_keys_without_ann)
        all_header += [format_key("INFO", key) for key in info_keys_without_ann]
        all_expression += ", " + info_expr

    format_keys = list(header.formats.keys())
    if format_keys:
        all_header += [format_key("FORMAT", key) for key in format_keys]
        format_expr = ", ".join(
            f'(FORMAT["{key}"] or {{}}).get(SAMPLE, NA)' for key in format_keys
        )
        all_expression += ", " + format_expr

    annotation_keys = get_annotation_keys(header, ann_key)
    if annotation_keys:
        annotation_expr = ", ".join(f'ANN["{key}"]' for key in annotation_keys)
        all_header += [format_key(ann_key, key) for key in annotation_keys]
        all_expression += ", " + annotation_expr

    return all_expression, all_header


def generate_for_each_sample_expressions(s: str, vcf: VCFReader) -> list[str]:
    from asttokens.util import replace

    # parse the `for_each_sample(lambda var: inner) expression
    var, inner = _var_and_body(s)

    # here we parse the inner part again, so that position offsets start at 0
    tok = asttokens.ASTTokens(inner, parse=True)

    # *replace* `var` with each value from the list of samples
    samples = list(vcf.header.samples)
    expanded = []
    for sample in samples:
        replacements = []
        for t in tok.tokens:
            if t.type == 1 and t.string == var:
                pos = (t.startpos, t.endpos)
                replacements.append((*pos, f"'{sample}'"))
        expanded.append(replace(inner, replacements))
    return expanded


def generate_for_each_sample_column_names(s: str, vcf: VCFReader) -> list[str]:
    # parse the `for_each_sample(lambda var: inner) expression
    var, inner = _var_and_body(s)

    samples = list(vcf.header.samples)
    __globals = default_allowed_globals.copy()

    column_names = []
    for sample in samples:
        __globals[var] = sample
        column_name = eval(inner, __globals, {})
        if not isinstance(column_name, str | bytes):
            if hasattr(column_name, "__str__"):
                column_name = str(column_name)
            else:
                raise ValueError(
                    "The specified header expression does not evaluate to a string."
                    "Consider using `str(expression)` instead.",
                )
        column_names.append(column_name)
    return column_names


def _var_and_body(s):
    import ast

    tok = asttokens.ASTTokens(s, parse=True)
    tok.mark_tokens(tok.tree)
    var = None
    inner = ""

    # for_each_sample must be the top level function
    # since that is the only place it is allowed in order to split the header expression
    # such that there is one column per sample (instead of *one* list-valued column)

    # walk the resulting AST, find the "for_each_sample" ast.Call node,
    # and extract the variable name and lambda body code from that
    for node in asttokens.util.walk(tok.tree):
        if isinstance(node, ast.Call) and hasattr(node.func, "id"):
            if node.func.id == "for_each_sample":
                for arg in node.args:
                    if isinstance(arg, ast.Lambda):
                        var = tok.get_text(arg.args)
                        inner = tok.get_text(arg.body)
    return var, inner


def preprocess_expression(
    expression: str,
    vcf: VCFReader,
    make_expression: bool = True,
) -> str:
    """
    Split the header expression at toplevel commas into parts.
    Then, if one of these parts starts with 'for_each_sample',
    that part is expanded for each sample in vcf.header.samples
    """

    if expression == ALL_EXPRESSION:
        return expression

    parts: list[str] = get_toplevel(expression)
    to_expand = list(
        filter(lambda x: x[1].startswith("for_each_sample"), enumerate(parts)),
    )
    if len(to_expand) > 0 and vcf is None:
        raise ValueError("If FORMAT is to be expanded, the VCF kwarg must not be none.")
    parts_expanded: list[list[str]] = [[p] for p in parts]
    func = (
        generate_for_each_sample_expressions
        if make_expression
        else generate_for_each_sample_column_names
    )
    for i, p in to_expand:
        expanded = func(p, vcf)
        parts_expanded[i] = expanded
    parts_flattened = [p for pp in parts_expanded for p in pp]
    return ", ".join(parts_flattened)


def get_header(args, vcf: VCFReader) -> list[str]:
    if args.expression == ALL_EXPRESSION:
        _, all_header = construct_all_expression_and_header(
            vcf.header, args.annotation_key, args.naming_convention
        )
        return all_header
    header = args.expression if args.header == "auto" else args.header
    if not args.wide:
        header = f"SAMPLE, {header}"
    return get_toplevel(preprocess_expression(header, vcf, args.header == "auto"))


def get_toplevel(header: str) -> list[str]:
    splitpos = [0]
    level = 0
    stack = []
    for i, c in enumerate(header):
        if c == "," and level == 0:
            splitpos.append(i + 1)
        elif c == "(":
            stack.append("(")
            level += 1
        elif c == ")":
            previous = stack[-1]
            if previous == "(":
                # found matching bracket pair
                stack.pop()
                level -= 1
            else:
                raise SyntaxError("No matching ( found.")
    if len(stack) != 0:
        raise SyntaxError("Imbalanced number of brackets.")
    splitpos.append(len(header) + 1)
    parts = []
    for start, end in zip(
        splitpos,
        splitpos[1:],
        strict=False,
    ):
        # remove leading + trailing whitespace
        parts.append(header[start : end - 1].strip())
    return parts


def get_row(row):
    if not isinstance(row, tuple):
        row = (row,)
    return row


@handle_vembrane_error
def execute(args):
    aux = read_auxiliary(args.aux)
    ontology = read_ontology(args.ontology)
    overwrite_number = {
        "INFO": dict(args.overwrite_number_info),
        "FORMAT": dict(args.overwrite_number_format),
    }
    with create_reader(
        args.vcf,
        backend=args.backend,
        overwrite_number=overwrite_number,
    ) as vcf:
        expression = preprocess_expression(args.expression, vcf, True)
        if expression == ALL_EXPRESSION:
            args.wide = False
            expression, _header = construct_all_expression_and_header(
                vcf.header, args.annotation_key, args.naming_convention
            )
        else:
            if not args.wide:
                if list(vcf.header.samples):
                    expression = f"SAMPLE, {expression}"
                else:
                    # if there are no samples, SAMPLE will be undefined;
                    # add an empty string instead
                    expression = f"'', {expression}"
        rows = tableize_vcf(
            vcf,
            expression,
            args.annotation_key,
            overwrite_number=overwrite_number,
            wide=args.wide,
            auxiliary=aux,
            auxiliary_globals=Context.from_args(args).get_globals(),
            ontology=ontology,
        )

        with smart_open(args.output, "wt", newline="") as csvfile:
            writer = csv.writer(
                csvfile,
                delimiter=args.separator,
                quoting=csv.QUOTE_MINIMAL,
            )
            if args.header != "none":
                header = get_header(args, vcf)
                n_header_cols = len(header)
                expr_cols = get_toplevel(expression)
                n_expr_cols = len(expr_cols)
                if n_header_cols != n_expr_cols:
                    raise HeaderWrongColumnNumberError(
                        n_expr_cols,
                        expr_cols,
                        n_header_cols,
                        header,
                    )
                writer.writerow(header)
            writer.writerows(get_row(row) for row in rows)



from collections import OrderedDict, defaultdict
from pathlib import Path
from sys import stderr
from typing import Dict, List, Optional, Tuple

import pysam
from pysam import VariantRecord
from pysam.libcbcf import VariantHeader

from vembrane.backend.base import (
    VCFHeader,
    VCFReader,
    VCFRecord,
    VCFRecordFilter,
    VCFRecordFormat,
    VCFRecordFormats,
    VCFRecordInfo,
    VCFWriter,
)

from ..ann_types import NA, type_info
from ..errors import UnknownInfoFieldError, UnknownSampleError


class PysamRecord(VCFRecord):
    def __init__(self, record: VariantRecord, record_idx: int, header: VCFHeader):
        super().__init__(record, record_idx, header)

    @property
    def contig(self) -> str:
        return self._raw_record.chrom

    @property
    def position(self) -> int:
        return self._raw_record.pos

    @property
    def stop(self) -> int:
        return self._raw_record.stop

    @property
    def id(self) -> str:
        return self._raw_record.id

    @property
    def reference_allele(self) -> str:
        return self._raw_record.ref

    @property
    def alt_alleles(self) -> Tuple[str]:
        return tuple(self._raw_record.alts)

    @property
    def quality(self) -> float:
        return self._raw_record.qual

    @property
    def filter(self) -> VCFRecordFilter:
        return PysamRecordFilter(self._raw_record)

    @property
    def info(self) -> VCFRecordInfo:
        return PysamRecordInfo(self)

    @property
    def formats(self) -> VCFRecordFormats:
        return PysamRecordFormats(self)

    @property
    def samples(self):
        return self._header.samples

    @property
    def header(self) -> VCFHeader:
        return self._header


class PysamRecordFormats(VCFRecordFormats):
    __slots__ = "_record"

    def __init__(self, record: PysamRecord):
        self._record = record

    def __getitem__(self, key):
        return PysamRecordFormat(key, self._record)

    def keys(self):
        return self._record._header.formats.keys()


class PysamRecordFormat(VCFRecordFormat):
    __slots__ = ("_format_key", "_record", "_header", "_raw_record")

    def __init__(
        self,
        format_key: str,
        record: PysamRecord,
    ):
        self._format_key = format_key
        self._record = record
        self._header = record._header
        self._raw_record: VariantRecord = record._raw_record

    def __getitem__(self, sample):
        if not self.__contains__(sample):
            raise UnknownSampleError(self.record, sample)
        meta = self._header.formats[self._format_key]
        number = meta["Number"] if self._format_key != "GT" else "."
        value_array = self._raw_record.samples[sample]

        try:
            value = value_array[self._format_key]
            return type_info(value, number)
        except KeyError:
            print(
                f"Warning: "
                f"record {self._record.record_idx} is missing a value "
                f"for FORMAT key {self._format_key}, "
                f"returning NA instead."
                f"\n{self._record}\n",
                file=stderr,
            )
            return type_info(NA, number)

    def __setitem__(self, key, value):
        self._raw_record.format[key] = value

    def __contains__(self, sample):
        return sample in self._header.samples


class PysamRecordInfo(VCFRecordInfo):
    __slots__ = ("_record", "_raw_record")

    def __init__(
        self,
        record: PysamRecord,
    ):
        self._record = record
        self._raw_record: VariantRecord = record._raw_record

    def __getitem__(self, key):
        if key == "END":
            return self._record.end
        try:
            meta = self._record._header.infos[key]
        except KeyError as ke:
            raise UnknownInfoFieldError(self._record, key) from ke
        try:
            value = self._raw_record.info[key]
        except KeyError:
            print(
                f"Warning: "
                f"record {self._record.record_idx} is missing a value "
                f"for INFO key {key}, "
                f"returning NA instead."
                f"\n{self._record}\n",
                file=stderr,
            )
            return type_info(NA, meta["Number"])
        return type_info(value, meta["Number"])

    def __setitem__(self, key, value):
        self._raw_record.info[key] = value

    def __contains__(self, key):
        return key in self._raw_record.info

    def keys(self):
        return self._raw_record.info.keys()


class PysamRecordFilter(VCFRecordFilter):
    __slots__ = "_record"

    def __init__(self, record: VariantRecord):
        self._record = record

    def __iter__(self):
        yield from self._record.filter

    def add(self, tag: str):
        self._record.filter.add(tag)


class PysamReader(VCFReader):
    __slots__ = (
        "filename",
        "_iter_file",
        "_header",
        "_current_record_idx",
    )

    def __init__(
        self,
        filename: str | Path,
        overwrite_number: Dict[str, Dict[str, str]] | None = None,
    ):
        if overwrite_number is None:
            overwrite_number = {}
        self.filename = filename
        self._file = pysam.VariantFile(str(self.filename))
        self._header = PysamHeader(self._file.header, overwrite_number)
        self._current_record_idx = 0

    def __iter__(self):
        self._iter_file = self._file.__iter__()
        return self

    def __next__(self):
        self._current_record_idx += 1
        return PysamRecord(
            self._iter_file.__next__(),
            self._current_record_idx,
            self._header,
        )

    def reset(self):
        self._current_record_idx = 0
        self._file.reset()


class PysamHeader(VCFHeader):
    __slots__ = (
        "_record",
        "_raw_header",
        "_metadata",
        "_metadata_category",
        "_metadata_generic",
        "_samples",
    )

    def __init__(self, header: VariantHeader, overwrite_number=None):
        if overwrite_number is None:
            overwrite_number = {}
        self._raw_header = header
        self._metadata = []
        self._metadata_category: defaultdict[str | None, OrderedDict] = defaultdict(
            OrderedDict
        )
        self._metadata_generic = dict()
        self._samples = OrderedDict((s, None) for s in self._raw_header.samples)

        for r in self._raw_header.records:
            if r.type == "GENERIC":
                self._metadata_generic[r.key] = r.value
                continue
            d = dict(r)
            self._metadata.append(d)
            if "ID" in d:
                self._metadata_category[r.type][d["ID"]] = d

        # override numbers
        for category, items in overwrite_number.items():
            for key, value in items.items():
                if key in self._metadata_category[category]:
                    self._metadata_category[category][key]["Number"] = value

    def contains_generic(self, key: str):
        return key in self._metadata_generic

    def get_generic(self, key: str):
        return self._metadata_generic[key]

    @property
    def infos(self):
        return self._metadata_category["INFO"]

    @property
    def formats(self):
        return self._metadata_category["FORMAT"]

    def __iter__(self):
        self._iter_index = 0
        return self

    def __next__(self):
        if self._iter_index >= len(self._metadata):
            raise StopIteration
        ret = self._metadata[self._iter_index]
        self._iter_index += 1
        return ret

    @property
    def samples(self):
        return self._samples

    @property
    def filters(self):
        return self._raw_header.filters

    @property
    def records(self):
        return self._metadata_category

    def add_generic(self, key: str, value: str):
        self._raw_header.add_meta(key, value)

    def add_filter(self, id: str, description: str):
        self._raw_header.add_meta(
            key="FILTER",
            items=[("ID", id), ("Description", description)],
        )

    def add_meta(
        self,
        key: str,
        value: Optional[str] = None,
        items: Optional[List[Tuple[str, str]]] = None,
    ):
        raise NotImplementedError


class PysamWriter(VCFWriter):
    __slots__ = ("filename", "_header", "_file")

    def __init__(self, filename: str | Path, fmt: str, template: VCFReader):
        self.filename = filename
        self._file = pysam.VariantFile(
            str(self.filename),
            f"w{fmt}",  # type: ignore
            header=template.header._raw_header,
        )
        self._header = PysamHeader(self._file.header)

    def write(self, record: VCFRecord):
        self._file.write(record._raw_record)



import csv
import importlib.resources
import sys
from collections import defaultdict
from typing import Any

# intervaltree is untyped, so we use type: ignore to suppress type checking errors
from intervaltree import IntervalTree  # type: ignore

from vembrane.common import (
    Context,
    HumanReadableDefaultsFormatter,
    Primitive,
    Singleton,
    add_common_arguments,
)
from vembrane.errors import VembraneError
from vembrane.globals import default_allowed_globals
from vembrane.modules import structured

PROFILE_DIR = (
    importlib.resources.files("vembrane.modules") / "assets" / "fhir" / "profiles"
)


def add_subcommand(subparsers):
    parser = subparsers.add_parser(
        "fhir",
        help="Generate FHIR records from VCF/BCF files.",
        description="Generate FHIR records from VCF/BCF files.",
        formatter_class=HumanReadableDefaultsFormatter,
    )
    parser.add_argument(
        "vcf",
        help="Path to the VCF/BCF file to be filtered. Defaults to '-' for stdin.",
        nargs="?",
        default="-",
    )
    parser.add_argument(
        "sample",
        help="The sample to use for generating FHIR output.",
    )
    parser.add_argument(
        "assembly",
        help="The reference assembly used for read mapping.",
        choices=Assemblies().names(),
    )
    parser.add_argument(
        "--url",
        "-u",
        help="Generic url used as identifier by FHIR e.g. "
        "http://<institute>/<department>/VCF",
        default=None,
    )
    parser.add_argument(
        "--status",
        "-s",
        help="Status of findings. E.g. final, preliminary, ...",
        default=None,
    )
    parser.add_argument(
        "--profile",
        required=True,
        choices=[path.with_suffix("").name for path in PROFILE_DIR.iterdir()],
        help="The FHIR profile to use for generating the output, see "
        "https://github.com/vembrane/vembrane/tree/main/vembrane/modules/assets/"
        "fhir/profiles for available profiles and the degree of support.",
    )
    parser.add_argument(
        "--id-source",
        help="URL to the source of IDs found in the ID column of the VCF/BCF file. "
        "IDs are only used if this is given.",
    )
    parser.add_argument(
        "--genomic-source-class",
        help="The genomic source class of the given variants as defined by "
        "LOINC: https://loinc.org/48002-0. Either provide the name as a string or "
        "a Python expression that evaluates to the name, e.g., for Varlociraptor "
        '\'"Somatic" if INFO["PROB_SOMATIC"] > 0.95 else ...\'.',
    )
    parser.add_argument(
        "--sample-allelic-frequency",
        default=None,
        help="Python expression calculating the the samples allelic frequency"
        "as percentage. E.g. \"FORMAT['AF'][sample][0] * 100\" ",
    )
    parser.add_argument(
        "--sample-allelic-read-depth",
        default="FORMAT['AD'][sample][1]",
        help="Python expression accessing the the samples allelic read depth."
        "Default is: \"FORMAT['AD'][sample][1]\"",
    )
    parser.add_argument(
        "--confidence-status",
        default=None,
        help="Python expression for calculating the variants confidence status being "
        "High, Intermediate or Low. E.g. \"'High' if QUAL >= 20 else ('Intermediate' "
        "if QUAL >= 10 else 'Low')\"",
    )
    parser.add_argument(
        "--detection-limit",
        type=int,
        help="Detection limit / sensitivity of the analysis in percent (e.g. 95).",
    )
    parser.add_argument(
        "--output-fmt",
        choices=["json", "jsonl", "yaml"],
        help="Output format. If not specified, can be automatically determined from "
        "the --output file extension.",
    )
    parser.add_argument(
        "--output",
        "-o",
        default="-",
        help="Output file, if not specified, output is written to STDOUT.",
    )
    add_common_arguments(parser)


def load_tsv(path, skip_comments=True):
    with open(path, "r") as file:
        reader = csv.reader(file, delimiter="\t")
        if skip_comments:
            return [line for line in reader if not line[0].startswith("#")]
        else:
            return list(reader)


class Cytobands(metaclass=Singleton):
    def __init__(self):
        self._data = defaultdict(IntervalTree)
        records = load_tsv(
            importlib.resources.files("vembrane.modules")
            / "assets"
            / "fhir"
            / "cytobands.tsv"
        )
        for chrom, start, end, band, _ in records:
            self._data[chrom].addi(int(start), int(end), band)

    def get(self, chrom: str, pos: int) -> str | None:
        if not chrom.startswith("chr"):
            chrom = f"chr{chrom}"
        intervals = self._data.get(chrom)
        if intervals:
            hits = intervals[pos]
            if hits:
                return next(iter(hits)).data
        return None


class Assemblies(metaclass=Singleton):
    def __init__(self):
        self._data = {
            assembly: loinc
            for assembly, loinc in load_tsv(
                importlib.resources.files("vembrane.modules")
                / "assets"
                / "fhir"
                / "assemblies.tsv",
                skip_comments=False,
            )
        }

    def get(self, assembly: str) -> str | None:
        return self._data.get(assembly)

    def names(self) -> list[str]:
        """
        Get a list of all assembly names.
        """
        return list(self._data.keys())


class Chromosomes(metaclass=Singleton):
    def __init__(self):
        self._data = {
            chrom: (chrom_display, loinc)
            for chrom, chrom_display, loinc in load_tsv(
                importlib.resources.files("vembrane.modules")
                / "assets"
                / "fhir"
                / "chromosomes.tsv",
                skip_comments=False,
            )
        }

    def get(self, chrom: str) -> tuple[str, str] | None:
        if not chrom.startswith("chr"):
            chrom = f"chr{chrom}"
        return self._data.get(chrom)


class GenomicSourceClasses(metaclass=Singleton):
    def __init__(self):
        self._data = {
            name: code
            for name, code in load_tsv(
                importlib.resources.files("vembrane.modules")
                / "assets"
                / "fhir"
                / "genomic_source_classes.tsv"
            )
        }

    def code(self, name: str | None) -> str | None:
        """
        Get the LOINC code for the given genomic source class name.
        """
        if name is None:
            return None
        return self._data[name]

    def is_name(self, expr: str | None) -> bool:
        """
        Check if the given expression is a valid genomic source class name.
        """
        if expr is None:
            return False
        return expr in self._data

    def preprocess_expression(self, expr: str | None) -> str | None:
        """
        Preprocess the expression to ensure it is a valid genomic source class name.
        If the expression is a string, it will be wrapped in quotes.
        """
        if expr is None:
            return None
        if self.is_name(expr):
            return f'"{expr}"'
        return expr


def execute(args):
    try:
        if args.detection_limit is not None:
            if not (0 < args.detection_limit <= 100):
                raise ValueError("Detection limit must be between 1 and 100.")

        genomic_source_classes = GenomicSourceClasses()

        with open(PROFILE_DIR / f"{args.profile}.yaml", mode="r") as infile:
            template = infile.read()

        structured.process(
            template=template,
            output_fmt=args.output_fmt,
            output=args.output,
            vcf=args.vcf,
            annotation_key=args.annotation_key,
            overwrite_number_info=args.overwrite_number_info,
            overwrite_number_format=args.overwrite_number_format,
            backend=args.backend,
            allowed_globals=default_allowed_globals
            | {
                "sample": args.sample,
                "url": args.url,
                "status": args.status,
                "assembly": args.assembly,
                "sample_allelic_frequency": args.sample_allelic_frequency,
                "sample_allelic_read_depth": args.sample_allelic_read_depth,
                "confidence_status": args.confidence_status,
                "cytobands": Cytobands(),
                "assemblies": Assemblies(),
                "chromosomes": Chromosomes(),
                "genomic_source_classes": genomic_source_classes,
                "genomic_source_class": genomic_source_classes.preprocess_expression(
                    args.genomic_source_class
                ),
                "id_source": args.id_source,
                "detection_limit": args.detection_limit,
            },
            auxiliary_globals=Context.from_args(args).get_globals(),
            postprocess=postprocess_fhir_record,
        )
    except VembraneError as ve:
        print(ve, file=sys.stderr)
        sys.exit(1)


def postprocess_fhir_record(record: Primitive | dict | list) -> Primitive | dict | list:
    assert isinstance(record, dict), "bug: FHIR record is expected to be a dictionary"

    def is_valid(entry: Any) -> bool:
        if isinstance(entry, dict):
            return all(is_valid(value) for value in entry.values())
        elif isinstance(entry, list):
            return all(is_valid(item) for item in entry)
        else:
            return entry is not None

    record["component"] = list(filter(is_valid, record["component"]))
    return record



from os import PathLike

try:
    from typing import NewType, Self, TextIO
except ImportError:
    from typing_extensions import NewType, Self, TextIO

import networkx
import obonet

DEFAULT_SEQUENCE_ONTOLOGY_FILE = "sequence_ontology.2024-06-06.obo.xz"

Id = NewType("Id", str)

_missing = object()


class CachedProperty(object):
    """
    A decorator for caching the result of a property.
    Replaces itself with the cached value after the first call.

    Inspired by https://github.com/pydanny/cached-property / django's cached_property
    """

    def __init__(self, func):
        self.__name__ = func.__name__
        self.func = func

    def __get__(self, obj, cls=None):
        if obj is None:
            return self
        value = obj.__dict__.get(self.__name__, _missing)
        if value is _missing:
            value = self.func(obj)
            obj.__dict__[self.__name__] = value
        return value


cached_property = CachedProperty


class _Container:
    """
    A container for the Sequence Ontology and related objects.
    This enables lazy-loading of the ontology.
    """

    ontology: "SequenceOntology"

    @cached_property  # type: ignore
    def ontology(self) -> "SequenceOntology":
        return SequenceOntology.default()


_C: _Container = _Container()


class Term(str):
    """
    Representation of a Sequence Ontology term.
    Inherits from `str` and adds methods for querying the ontology.
    """

    def get_id(self) -> Id:
        """Get the ID of this term."""
        try:
            return _C.ontology.term_to_id[self]
        except KeyError as ke:
            raise ValueError(f"Term {self} not found in the ontology") from ke

    def parents(self) -> "Consequences":
        """Return the parent terms of this term."""
        # obo file has reversed edges,
        # so networkx descendants are ancestors and vice versa
        return Consequences(
            map(_C.ontology.get_term, _C.ontology.graph.successors(self.get_id()))
        )

    def ancestors(self) -> "Consequences":
        """
        Return the ancestor terms of this term,
        i.e. from the term to the root.
        (Parents, Grandparents, etc.)
        """
        # obo file has reversed edges,
        # so networkx descendants are ancestors and vice versa
        return Consequences(
            map(
                _C.ontology.get_term,
                networkx.descendants(_C.ontology.graph, self.get_id()),
            )
        )

    def children(self) -> "Consequences":
        """Return the child terms of this term."""
        # obo file has reversed edges,
        # so networkx ancestors are descendants and vice versa
        return Consequences(
            map(_C.ontology.get_term, _C.ontology.graph.predecessors(self.get_id()))
        )

    def descendants(self) -> "Consequences":
        """
        Return the descendant terms of this term,
        i.e. from the term to the leaves.
        (Children, Grandchildren, etc.)
        """
        # obo file has reversed edges,
        # so networkx ancestors are descendants and vice versa
        return Consequences(
            map(
                _C.ontology.get_term,
                networkx.ancestors(_C.ontology.graph, self.get_id()),
            )
        )

    def is_ancestor(self, target_term: Self) -> bool:
        """
        Check whether this term is an ancestor of the target term.
        """
        if not target_term:
            return False
        return _C.ontology.path_lengths.get(target_term, {}).get(self) is not None

    def is_descendant(self, target_term: Self) -> bool:
        """
        Check whether this term is a descendant of the target term.
        """
        if not target_term:
            return False
        return _C.ontology.path_lengths.get(self, {}).get(target_term) is not None

    def is_a(self, target_term: Self) -> bool:
        """
        Check whether this term is a descendant of the target term.

        This delegates to `is_descendant`, i.e. is just a synonym for convenience.
        """
        return self.is_descendant(target_term)

    def path_length(self, target_term: Self) -> int | None:
        """
        Return the path length between this term and the target term.

        This is irrespective of the direction of the path.
        Returns `None` if there is no path between the terms.
        """
        return _C.ontology.path_lengths.get(self, {}).get(
            target_term
        ) or _C.ontology.path_lengths.get(target_term, {}).get(self)


class Consequences(list[Term]):
    def any_is_a(self, target_term: Term) -> bool:
        return any(term.is_a(target_term) for term in self)

    def most_specific_terms(self) -> "Consequences":
        """
        Return the most specific terms from the given list of terms.
        """
        return Consequences(
            term
            for term in self
            if not any(other.is_descendant(term) for other in self if term != other)
        )

    def __eq__(self, other):
        if isinstance(other, Consequences) or isinstance(other, list):
            return super().__eq__(other)
        else:
            raise TypeError(
                "Can only directly compare Consequences to other Consequences "
                "or lists of Terms, not str or any other types.\n"
                "If you want to check "
                "whether a single term is part of the Consequences, "
                "use the `in` operator, "
                "e.g. `'missense_variant' in ANN['Consequences']`.\n"
                "If you do want an exact match of a list of terms, "
                "use e.g. `['missense_variant'] == ANN['Consequences']`.\n"
                "Otherwise, prefer using the `any_is_a` method,"
                "e.g. `ANN['Consequences'].any_is_a('missense_variant')`."
            )

    def __str__(self) -> str:
        return "&".join(self)


class SequenceOntology:
    """
    Representation of the Sequence Ontology.
    Contains the graph, term to id mapping, id to term mapping,
    as well as pre-computed path lengths between all terms.
    """

    graph: networkx.MultiDiGraph
    term_to_id: dict[Term, Id]
    id_to_term: dict[Id, Term]
    path_lengths: dict[Term, dict[Term, int]]

    def __init__(self, graph: networkx.MultiDiGraph):
        self.graph = graph
        self.term_to_id = {
            Term(data["name"]): id_ for id_, data in graph.nodes(data=True)
        }
        self.id_to_term = {v: k for k, v in self.term_to_id.items()}
        self.path_lengths = {
            self.id_to_term[source_id]: {
                self.id_to_term[target_id]: length
                for target_id, length in lengths.items()
            }
            for source_id, lengths in networkx.all_pairs_shortest_path_length(graph)
        }

    @classmethod
    def from_obo(cls, path: str | PathLike | TextIO):
        """Read an Ontology from an OBO file."""
        if isinstance(path, list) and len(path) == 1:
            graph = obonet.read_obo(path[0])
        elif isinstance(path, (str, PathLike, TextIO)):
            graph = obonet.read_obo(path)
        else:
            raise TypeError(
                f"Expected str, PathLike, or TextIO, got {type(path)} for value {path}"
            )
        return cls(graph)

    @classmethod
    def default(cls):
        """Load the default Sequence Ontology from bundled resources."""
        import importlib.resources

        so_resource = importlib.resources.files("vembrane.resources").joinpath(
            DEFAULT_SEQUENCE_ONTOLOGY_FILE
        )
        with importlib.resources.as_file(so_resource) as so_file:
            return cls.from_obo(so_file)

    def get_id(self, term: Term) -> Id:
        """Get the ID of a term."""
        try:
            return self.term_to_id[term]
        except KeyError as ke:
            raise ValueError(f"Term {term} not found in the ontology") from ke

    def get_term(self, id_: Id) -> Term:
        """Get the term for an ID."""
        try:
            return self.id_to_term[id_]
        except KeyError as ke:
            raise ValueError(f"ID {id_} not found in the ontology") from ke



import heapq
import math
import sys
import tempfile
from functools import partial, total_ordering
from itertools import batched
from pathlib import Path
from typing import Any, Callable, Dict, Iterable, List, Self, Sequence, Tuple, TypeVar

from vembrane import __version__
from vembrane.ann_types import NA
from vembrane.backend.base import Backend, VCFReader, VCFRecord, VCFWriter
from vembrane.common import (
    Context,
    HumanReadableDefaultsFormatter,
    add_common_arguments,
    check_expression,
    create_reader,
    create_writer,
    normalize,
    read_auxiliary,
)
from vembrane.errors import VembraneError
from vembrane.representations import Annotation, SourceEnvironment


def add_subcommand(subparsers):
    parser = subparsers.add_parser(
        "sort",
        help="Sort VCF/BCF records by one or multiple Python expressions.",
        description="Sort VCF/BCF records by one or multiple Python expressions that "
        "encode keys for the desired order. This feature is primarily meant to "
        "prioritizing records for the human eye. For large unfiltered VCF/BCF files, "
        "the only relevant sorting is usually by position, which is better done with "
        "e.g. bcftools (and usually the standard sorting that variant callers output). "
        "Expects the VCF/BCF file to be single-allelic, i.e., one ALT allele per "
        "record.",
        formatter_class=HumanReadableDefaultsFormatter,
    )
    parser.add_argument(
        "expression",
        type=check_expression,
        help="Python expression (or tuple of expressions) returning orderable values "
        "(keys) to sort the VCF/BCF records by. "
        "By default keys are considered in ascending order. "
        "To sort by descending order, use `desc(<expression>)` on the entire "
        "expression or on individual items of the tuple. "
        "If multiple expressions are provided as a tuple, they are prioritized from "
        "left to right with lowest priority on the right. "
        "NA/NaN values are always sorted to the end. "
        "Expressions on annotation entries (there can be multiple annotations per "
        "record) will cause the annotation with the "
        "minimum key value (or maximum if descending) to be considered to sort "
        "the record.",
    )
    parser.add_argument(
        "vcf",
        help="The VCF/BCF file containing the variants. If not specified, "
        "reads from STDIN.",
        nargs="?",
        default="-",
    )
    parser.add_argument(
        "--output",
        "-o",
        default="-",
        help="Output file, if not specified, output is written to STDOUT.",
    )
    parser.add_argument(
        "--output-fmt",
        "-O",
        default="vcf",
        choices=["vcf", "bcf", "uncompressed-bcf"],
        help="Output format.",
    )
    parser.add_argument(
        "--preserve-annotation-order",
        action="store_true",
        help="If set, annotations are not sorted within the records, but kept in the "
        "same order as in the input VCF/BCF file. If not set (default), annotations "
        "are sorted within the record according to the given keys if any of the sort "
        "keys given in the python expression refers to an annotation.",
    )
    parser.add_argument(
        "--max-in-mem-records",
        type=int,
        default=100000,
        help="Number of VCF/BCF records to sort in memory. If the VCF/BCF file exceeds "
        "this number of records, external sorting is used.",
    )
    add_common_arguments(parser)


def is_na(value: Any) -> bool:
    return (
        value is NA or value is None or (isinstance(value, float) and math.isnan(value))
    )


T = TypeVar("T")

type NestedSequence[T] = Sequence[T | NestedSequence[T]]


# Key wrapper classes for sorting.
# sorted() uses <, max and min use < and >, so we need to implement both.
# For maximizing speed, we avoid delegation, saving an additional method call.
class KeyBase:
    @classmethod
    def wrap(cls, key: Any) -> Self | NestedSequence[Self]:
        if isinstance(key, Iterable) and not isinstance(key, str):
            # apply recursively to tuples, lists, etc.
            return tuple(cls.wrap(k) for k in key)
        else:
            return cls(key)

    def __init__(self, key: Any):
        self.key = key

    # rationale for the type ignore below: the code ensures that other is of same type
    # hence we avoid additional checks here for performance reasons.
    def __eq__(self, other: Self) -> bool:  # type: ignore
        if is_na(self.key) and is_na(other.key):
            return True
        if is_na(self.key) or is_na(other.key):
            return False
        return self.key == other.key


@total_ordering
class KeyAscending(KeyBase):
    def __lt__(self, other: Self) -> bool:
        if is_na(self.key):
            return False
        if is_na(other.key):
            return True
        return self.key < other.key

    def __gt__(self, other: Self) -> bool:
        if is_na(self.key):
            return True
        if is_na(other.key):
            return False
        return self.key > other.key


@total_ordering
class KeyDescending(KeyBase):
    def __lt__(self, other: Self) -> bool:
        if is_na(self.key):
            return False
        if is_na(other.key):
            return True
        return self.key > other.key

    def __gt__(self, other: Self) -> bool:
        if is_na(self.key):
            return True
        if is_na(other.key):
            return False
        return self.key < other.key


type SortKey = (
    KeyDescending | KeyAscending | NestedSequence[KeyDescending | KeyAscending]
)


def apply_default_key(
    key: Any,
) -> SortKey:
    if isinstance(key, (KeyDescending, KeyAscending)):
        return key
    else:
        return KeyAscending.wrap(key)


auxiliary_globals: dict[str, Any] = {
    "asc": KeyAscending.wrap,
    "desc": KeyDescending.wrap,
}


def execute(args) -> None:
    overwrite_number = {
        "INFO": dict(args.overwrite_number_info),
        "FORMAT": dict(args.overwrite_number_format),
    }
    fmt = {"vcf": "", "bcf": "b", "uncompressed-bcf": "u"}[args.output_fmt]

    aux = read_auxiliary(args.aux)

    try:
        with create_reader(
            args.vcf,
            backend=args.backend,
            overwrite_number=overwrite_number,
        ) as reader:
            # TODO: unify the reader/writer setup across subcommands
            reader.header.add_generic("vembraneVersion", __version__)
            reader.header.add_generic(
                "vembraneCmd",
                "vembrane "
                + " ".join(
                    normalize(arg) if " " in arg else arg for arg in sys.argv[1:]
                ),
            )

            annotation = Annotation(args.annotation_key, reader.header)

            sort_keys = SourceEnvironment(
                args.expression,
                args.annotation_key,
                reader.header,
                auxiliary_globals=Context.from_args(args).get_globals()
                | auxiliary_globals,
                auxiliary=aux,
            )

            def get_sort_key(
                item: Tuple[int, VCFRecord], sort_annotations: bool = False
            ) -> SortKey:
                idx, record = item
                sort_keys.update_from_record(idx, record)

                if sort_keys.expression_annotations():
                    # if the expression refers to annotations, we obtain the
                    # minimal keys across all annotations
                    annotations = annotation.get_record_annotations(idx, record)

                    def eval_with_ann(annotation):
                        # load given annotation into the eval context
                        sort_keys.update_annotation(annotation)
                        # eval key expression and ensure that default SortKey is applied
                        # if not explicitly annotated by the user
                        return apply_default_key(
                            eval(
                                sort_keys.compiled,
                                sort_keys,
                            )
                        )

                    if sort_annotations:
                        # obtain keys for all annotations
                        keys = list(map(eval_with_ann, annotations))
                        # sort by keys
                        sorted_idx = sorted(
                            range(len(annotations)), key=keys.__getitem__
                        )
                        # order annotations according to the sorted keys
                        sorted_annotations = [annotations[i] for i in sorted_idx]
                        # update annotations in the record
                        annotation.set_record_annotations(record, sorted_annotations)
                        # return min key (SortKey ensures that this implements
                        # either ascending or descending order)
                        return keys[sorted_idx[0]]
                    else:
                        # return min key (SortKey ensures that this implements
                        # either ascending or descending order)
                        return min(map(eval_with_ann, annotations))
                else:
                    return apply_default_key(eval(sort_keys.compiled, sort_keys))

            with create_writer(
                args.output, fmt, reader, backend=args.backend
            ) as writer:
                # perform the actual sorting
                external_sort(
                    reader,
                    writer,
                    key=get_sort_key,
                    chunk_size=args.max_in_mem_records,
                    backend=args.backend,
                    overwrite_number=overwrite_number,
                    preserve_annotation_order=args.preserve_annotation_order,
                )
    except VembraneError as ve:
        print(ve, file=sys.stderr)
        sys.exit(1)


def external_sort(
    reader: VCFReader,
    writer: VCFWriter,
    key: Callable[[Tuple[int, VCFRecord]], SortKey],
    chunk_size: int,
    backend: Backend,
    overwrite_number: Dict[str, Dict[str, str]],
    preserve_annotation_order: bool,
):
    # Open temp dir for storing sorted chunks in case of external sorting
    with tempfile.TemporaryDirectory() as temp_dir:
        temp_path = Path(temp_dir)
        chunk_files: List[Path] = []

        last_chunk = None

        def store_chunk(chunk):
            path = temp_path / f"chunk_{len(chunk_files)}.bcf"
            with create_writer(path, "", reader, backend=backend) as writer:
                for _, record in chunk:
                    writer.write(record)
            chunk_files.append(path)

        for chunk in batched(enumerate(reader), n=chunk_size):
            if last_chunk is not None:
                # if there was a previous chunk, store it now
                store_chunk(last_chunk)
            # sort by key function (SortKey ensures that this implements
            # either ascending or descending order)
            # What we do here is seemingly too much for mypy to understand.
            # But it is definitely correct and works fine.
            # Thus the type: ignore below.
            chunk = sorted(
                chunk,
                key=partial(key, sort_annotations=not preserve_annotation_order),  # type: ignore
            )  # type: ignore

            last_chunk = chunk
        if last_chunk:
            if not chunk_files:
                # only one chunk, nothing stored yet
                # directly write the sorted chunk to the output
                # Thus, no external sorting is needed!
                for _, record in last_chunk:
                    writer.write(record)
                return
            else:
                # store last chunk of the iteration
                store_chunk(last_chunk)
        else:
            # empty input
            return

        # now merge the chunks
        readers = {}
        for i, chunk_file in enumerate(chunk_files):
            # open a reader for each chunk
            reader = create_reader(
                chunk_file, backend=backend, overwrite_number=overwrite_number
            )
            readers[i] = enumerate(reader), reader

        def next_record_to_heap(reader_idx: int) -> None:
            reader_iter, reader = readers[reader_idx]
            try:
                # get the next record of the given reader
                idx, record = next(reader_iter)
            except StopIteration:
                # no more records, close reader
                reader.close()
                del readers[reader_idx]
                return
            # store the record in the heap
            heapq.heappush(heap, (key((idx, record)), record, reader_idx))

        heap: List[Tuple[SortKey, VCFRecord, int]] = []
        for reader_idx in range(len(readers)):
            next_record_to_heap(reader_idx)

        while True:
            # Pop the smallest item from the heap.
            # Since all chunks are sorted and each has given one item to the heap,
            # the smallest item is the next record to write.
            try:
                _, record, reader_idx = heapq.heappop(heap)
            except IndexError:
                # heap empty, everything has been processed
                assert not readers
                return

            writer.write(record)
            # Get the next record from the same reader and push it to the heap.
            # This ensures that the heap always contains the next smallest record
            # from each reader. The heap sorts it into the right place.
            next_record_to_heap(reader_idx)
